from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import default_reinterpret_cast_fn
from hwt.synthesizer.exceptions import TypeConversionErr
from hwt.code import Concat
from hwt.hdl.types.array import HArray
from hwt.bitmask import selectBitRange, mask


def getBits_from_array(array, wordWidth, start, end,
                       reinterpretElmToType=None):
    """
    Gets value of bits between selected range from memory

    :param start: bit address of start of bit of bits
    :param end: bit address of first bit behind bits
    :return: instance of BitsVal (derived from SimBits type) which contains
        copy of selected bits
    """
    inPartOffset = 0
    value = Bits(end - start, None).fromPy(None)

    while start != end:
        assert start < end, (start, end)

        dataWordIndex = start // wordWidth

        v = array[dataWordIndex]
        if reinterpretElmToType is not None:
            v = v._reinterpret_cast(reinterpretElmToType)

        endOfWord = (dataWordIndex + 1) * wordWidth
        width = min(end, endOfWord) - start
        offset = start % wordWidth

        val = selectBitRange(v.val, offset, width)
        vldMask = selectBitRange(v.vldMask, offset, width)
        updateTime = v.updateTime

        m = mask(width)
        value.val |= (val & m) << inPartOffset
        value.vldMask |= (vldMask & m) << inPartOffset
        value.updateMask = max(value.updateTime, updateTime)

        inPartOffset += width
        start += width

    return value


def reinterptet_harray_to_bits(typeFrom, sigOrVal, bitsT):
    """
    Cast HArray signal or value to signal or value of type Bits
    """
    size = int(typeFrom.size)
    widthOfElm = typeFrom.elmType.bit_length()
    w = bitsT.bit_length()
    if size * widthOfElm != w:
        raise TypeConversionErr(
            "Size of types is different", size * widthOfElm, w)

    partT = Bits(widthOfElm)
    parts = [p._reinterpret_cast(partT) for p in sigOrVal]

    return Concat(*reversed(parts))._reinterpret_cast(bitsT)


def reinterpret_harray_to_harray(typeFrom, sigOrVal, arrayT):
    mySize = int(typeFrom.size)
    myWidthOfElm = typeFrom.elmType.bit_length()
    size = int(arrayT.size)
    widthOfElm = arrayT.elmType.bit_length()

    if size * widthOfElm != mySize * myWidthOfElm:
        raise TypeConversionErr("Size of types is different",
                                size * widthOfElm, mySize * myWidthOfElm)

    if isinstance(typeFrom.elmType, Bits):
        reinterpretElmToType = None
    else:
        reinterpretElmToType = Bits(myWidthOfElm)

    res = arrayT.fromPy(None)
    for i in range(size):
        start = i * widthOfElm
        end = (i + 1) * widthOfElm
        item = getBits_from_array(
            sigOrVal, myWidthOfElm, start, end, reinterpretElmToType)
        res[i] = item._reinterpret_cast(arrayT.elmType)

    return res


def reinterpret_cast_harray(typeFrom, sigOrVal, toType):
    if isinstance(toType, Bits):
        return reinterptet_harray_to_bits(typeFrom, sigOrVal, toType)
    elif isinstance(toType, HArray):
        return reinterpret_harray_to_harray(typeFrom, sigOrVal, toType)
    else:
        return default_reinterpret_cast_fn(typeFrom, sigOrVal, toType)



from itertools import islice, zip_longest

from hwt.hdl.assignment import Assignment
from hwt.hdl.operator import Operator
from hwt.hdl.statements import IfContainer, SwitchContainer
from hwt.hdl.value import Value
from hwt.pyUtils.arrayQuery import areSetsIntersets, groupedby
from hwt.serializer.utils import maxStmId


def removeUnconnectedSignals(netlist):
    """
    If signal is not driving anything remove it
    """

    toDelete = set()
    toSearch = netlist.signals

    while toSearch:
        _toSearch = set()

        for sig in toSearch:
            if not sig.endpoints:
                try:
                    if sig._interface is not None:
                        # skip interfaces before we want to check them,
                        # they should not be optimized out from design
                        continue
                except AttributeError:
                    pass

                for e in sig.drivers:
                    # drivers of this signal are useless rm them
                    if isinstance(e, Operator):
                        for op in e.operands:
                            if not isinstance(op, Value):
                                try:
                                    op.endpoints.remove(e)
                                except KeyError:
                                    # this operator has 2x+ same operand
                                    continue

                                _toSearch.add(op)

                    elif isinstance(e, Assignment):
                        op = e.src
                        if not isinstance(op, Value):
                            op.endpoints.remove(e)
                            _toSearch.add(op)

                        netlist.startsOfDataPaths.remove(e)
                    else:
                        raise AssertionError("Drivers should be only"
                                             "index operators or assignments", e)

                toDelete.add(sig)

        if toDelete:
            for sig in toDelete:
                netlist.signals.remove(sig)
                try:
                    _toSearch.remove(sig)
                except KeyError:
                    pass
            toDelete = set()
        toSearch = _toSearch


class IncompatibleStructure(Exception):
    """
    instances of HWProcess can not be merged due incompatible structure
    """


def isMergableStmList(listA, listB):
    la = len(listA)
    lb = len(listB)
    if la == lb:
        return True
    elif la > 0 and lb == 1:
        return True
    else:
        return False


def checkIfIsTooSimple(proc):
    """check if process is just unconditional assignments
       and it is useless to merge them"""
    try:
        a, = proc.statements
        if isinstance(a, Assignment):
            return True
    except ValueError:
        pass
    return False


def mergeStmLists(stmsA, stmsB):
    tmp = []
    for a, b in zip_longest(stmsA, stmsB, fillvalue=None):
        if b is None:
            assert a is not None
            tmp.append(a)
        else:
            tmp.extend(tryToMergeStm(a, b))

    return tmp


def tryToMergeStm(stmA, stmB):
    """
    :raise IncompatibleStructure: if it is not possible to merge statements
    :return: generator of statements
    """

    if isinstance(stmA, Assignment) or isinstance(stmB, Assignment):
        yield stmA
        yield stmB
        return

    aIsIf = isinstance(stmA, IfContainer)
    bIsIf = isinstance(stmB, IfContainer)
    if aIsIf and bIsIf:
        if (stmA.cond != stmB.cond or
                not isMergableStmList(stmA.ifTrue, stmB.ifTrue) or
                not isMergableStmList(stmA.elIfs, stmB.elIfs) or
                not isMergableStmList(stmA.ifFalse, stmB.ifFalse)):
            raise IncompatibleStructure()

        ifTrue = mergeStmLists(stmA.ifTrue, stmB.ifTrue)

        elIfs = []
        for (condA, elifA), (condB, elifB) in zip(stmA.elIfs, stmB.elIfs):
            if (condA != condB or not isMergableStmList(elifA, elifB)):
                raise IncompatibleStructure()
            elIfs.append((condA, mergeStmLists(elifA, elifB)))

        ifFalse = mergeStmLists(stmA.ifFalse, stmB.ifFalse)

        yield IfContainer(stmA.cond, ifTrue, ifFalse, elIfs)
        return

    aIsSwitch = isinstance(stmA, SwitchContainer)
    bIsSwitch = isinstance(stmB, SwitchContainer)
    if aIsSwitch and bIsSwitch:
        if not (stmA.switchOn is stmB.switchOn and
                len(stmA.cases) == len(stmB.cases) and
                isMergableStmList(stmA.default, stmB.default)):
            raise IncompatibleStructure()

        cases = []
        for (vA, caseA), (vB, caseB) in zip(stmA.cases, stmB.cases):
            if vA != vB or not isMergableStmList(caseA, caseB):
                raise IncompatibleStructure()
            cases.append((vA, mergeStmLists(caseA, caseB)))

        default = mergeStmLists(stmA.default, stmB.default)

        yield SwitchContainer(stmA.switchOn, cases, default)
        return

    if (aIsSwitch and bIsIf) or (aIsIf and bIsSwitch):
        raise IncompatibleStructure()

    raise NotImplementedError(stmA, stmB)


def tryToMerge(procA, procB):
    """
    Try merge procB into procA

    :raise IncompatibleStructure: if merge is not possible
    :attention: procA is now result if merge has succeed
    :return: procA which is now result of merge
    """
    if (checkIfIsTooSimple(procA) or
            checkIfIsTooSimple(procB) or
            areSetsIntersets(procA.outputs, procB.sensitivityList) or
            areSetsIntersets(procB.outputs, procA.sensitivityList) or
            len(procA.statements) != len(procB.statements)):
        raise IncompatibleStructure()

    statements = []
    for stmA, stmB in zip(procA.statements, procB.statements):
        statements.extend(tryToMergeStm(stmA, stmB))

    procA.statements = statements
    procA.outputs.update(procB.outputs)
    procA.inputs.update(procB.inputs)
    procA.sensitivityList.update(procB.sensitivityList)

    return procA


def reduceProcesses(processes, procRanks):
    """
    Try to merge processes as much is possible

    :param processes: list of processes instances
    :param procRanks: process ranks = how many assignments is probably in process
        used to minimize number of merge tries
    """
    # sort to make order of merging same deterministic
    processes.sort(key=lambda x: (x.name, maxStmId(x)))
    # now try to reduce processes with nearly same structure of statements into one
    # to minimize number of processes
    for _, procs in groupedby(processes, lambda p: procRanks[p]):
        for iA, pA in enumerate(procs):
            if pA is None:
                continue
            for iB, pB in enumerate(islice(procs, iA + 1, None)):
                if pB is None:
                    continue
                try:
                    pA = tryToMerge(pA, pB)
                except IncompatibleStructure:
                    continue
                procs[iA + 1 + iB] = None
                # procs[iA] = pA

        for p in procs:
            if p is not None:
                yield p



from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase

intfMethods = ["_getAssociatedClk", "_getAssociatedRst", "_connectToIter"]
signalMethods = ["_convert"
                 "_onRisingEdge", "_onFallingEdge",
                 '__div__', '__floordiv__', '__mod__', '__mul__', '__truediv__',
                 "_isOn", "_eq", "__ne__", "__gt__", "__lt__", "__ge__", "__le__",
                 "__invert__", "__neg__", "__and__", "__xor__", "__or__", "__add__", "__sub__", "__mul__",
                 "_reversed", "_concat"]


def delegated_methods(methodNames):
    """
    Delegate methods on object under "propName" property
    """
    def add_method(op, cls):
        if op in intfMethods:
            def delegated_op(self, *args, **kwargs):
                """
                call function from interface with self=me
                """
                o = self._origIntf

                fn = getattr(o.__class__, op)
                return fn(self, *args, **kwargs)
        else:
            def delegated_op(self, *args, **kwargs):
                """
                call function from interface with self=me
                """
                try:
                    o = self._sig
                except AttributeError:
                    o = self._origIntf

                fn = getattr(o.__class__, op)
                return fn(self, *args, **kwargs)

        setattr(cls, op, delegated_op)

    def decorator(cls):
        for op in methodNames:
            add_method(op, cls)
        return cls

    return decorator


@delegated_methods(signalMethods + intfMethods)
class InterfaceProxy(InterfaceBase):
    """
    Interface proxy which is used to create virtual arrays on interfaces
    goal is to have arrays just by multiplying the width of signals
    to allow intuitive access by [] operator to such a array we need this proxy

    :ivar _interfaces: list of proxies on interfaces from origInterface
    :ivar offset: tells how many items of this type was before this item f.e.
        arr = [[ item for i in range(3)] for i in range(3)] for array[2][0] index=0 and offset=2*3
    :ivar _itemsCnt: if this is an proxy for array this is size of this array
    :ivar _itemsInOne: data width multiplier for this proxy
    """

    def __init__(self, origInterface, offset, index, itemsCnt, itemsInOne):
        """
        :param origInterface: interface which is this proxy created for
        :param offset: how many of this items items have appeared in previous arrays (take a look at doc of class)
        :param index: index in current array of proxies, use None when parent proxy is not an array
        :param itemsCnt: if this is array specify how many items is in int
        """
        # if parentProxy is None this is constructed as element for _arrElemCache
        assert not isinstance(origInterface, InterfaceProxy)
        self._origIntf = origInterface
        self._index = index
        self._offset = offset
        self._itemsCnt = itemsCnt
        self._arrayElemCache = []
        assert itemsInOne > 0
        self._itemsInOne = itemsInOne

        self._interfaces = []
        for intf in origInterface._interfaces:
            _itemsCnt = intf._asArraySize
            if _itemsCnt is not None:
                _itemsCnt = int(_itemsCnt)

            if itemsCnt is not None:
                _itemsInOne = itemsInOne * itemsCnt
            else:
                _itemsInOne = itemsInOne

            p = InterfaceProxy(intf, offset, None, _itemsCnt, _itemsInOne)
            setattr(self, intf._name, p)
            self._interfaces.append(p)

        if itemsCnt is not None:
            self._initArrayItems()
        
    def __len__(self):
        """
        :return: number of items in this array proxy (they are proxy too)
        """
        c = self._itemsCnt
        if c is None:
            raise TypeError("This is not an array")

        return c

    def _myArrOffset(self):
        """
        Returns index in items on physical interface which corresponds to signals of this proxy
        """
        return self._offset

    def _isInterfaceArray(self):
        return self._itemsCnt is not None

    def _getMySigSelector(self):
        """
        :return: tuple (upper, lower) for selecting signal for this proxy from parent
            signal
        """
        w = self._origIntf._widthMultiplier
        assert w is not None, ("InterfceProxy is not expected to be on interfaces without multiplier", self)
        sigItemsCnt = int(w)

        width = self._origIntf._dtype.bit_length()
        widthOfItem = (width // sigItemsCnt) * self._itemsInOne
        index = self._myArrOffset()

        if widthOfItem == 1:  # [FIXME] it is not sure that type was originally bit or vector of len=1
            # as single bit it was single bit in original interface
            return index
        else:
            h = ((index + 1) * widthOfItem)
            l = (index * widthOfItem)
            # assert h <= width, self
            return (h, l)

    def _signalsForInterface(self, context):
        """
        Set _sig property for all signals recursively as a select from parent signals

        :param context: instance of RtlNetlist where signals should be created
        """

        if not self._interfaces:
            # this is proxy for signal interface, select bits from it as _sig of this proxy
            s = self._origIntf._sig
            index = self._getMySigSelector()
            if isinstance(index, tuple):
                h, l = index
                self._sig = s[h:l]
            else:
                self._sig = s[index]
        else:
            # there we know that all are proxies and all are part of array with dimension
            # specified in this Proxy or it's  parents
            for proxy in self._interfaces:
                proxy._signalsForInterface(context)

        for proxy in self._arrayElemCache:
            proxy._signalsForInterface(context)

    def _clean(self, rmConnetions=True, lockNonExternal=True):
        """
        Remove all signals from this interface (used after unit is synthesized
        and its parent is connecting its interface to this unit)
        """
        if self._interfaces:
            for i in self._interfaces:
                i._clean(rmConnetions=rmConnetions, lockNonExternal=lockNonExternal)
        else:
            self._sigInside = self._sig
            del self._sig

        for e in self._arrayElemCache:
            e._clean(rmConnetions=rmConnetions, lockNonExternal=lockNonExternal)

    def _connectTo(self, master, exclude=None, fit=False):
        """
        connect to another interface interface (on rtl level)
        works like self <= master in VHDL
        """
        return list(self._connectToIter(master, exclude, fit))

    def __call__(self, other):
        """
        :attention: it is not call of function it is operator of assignment
        """
        return self._connectTo(other)

    def __getitem__(self, key):
        return self._arrayElemCache[key]

    def __getattr__(self, name):
        if name == "_getIndexCascade" or name == "naryOp" or name == "_dtype":
            # called on _sig to avoid inf. loop
            o = self._sig
        elif name == "_initSimAgent":
            return lambda : self._origIntf.__class__._initSimAgent(self)
        else:
            o = self._origIntf

        return getattr(o, name)

    def _initArrayItems(self):
        "instantiate my items into _arrayElemCache"
        # I am the array proxy and now I instantiate proxies for my elements
        offset = self._myArrOffset() * self._itemsCnt

        for index in range(len(self)):
            itemsInOne = self._itemsInOne
            if self._itemsInOne is not None:
                itemsInOne = self._itemsInOne * itemsInOne

            e = InterfaceProxy(self._origIntf, offset + index, index, None, itemsInOne)
            self._arrayElemCache.append(e)

    def __repr__(self):
        return "<InterfaceProxy for %r, itemsInOne:%d, itemsCnt=%r, offset=%r, index=%r>" % (self._origIntf, self._itemsInOne, self._itemsCnt, self._offset, self._index)



import re

from hwt.hdl.assignment import Assignment
from hwt.hdl.entity import Entity
from hwt.hdl.types.array import HArray
from hwt.hdl.types.enum import HEnum
from hwt.hdl.variables import SignalItem
from hwt.pyUtils.arrayQuery import groupedby
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.serializer import GenericSerializer
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.generic.mapExpr import MapExpr
from hwt.serializer.generic.nameScope import LangueKeyword, NameScope
from hwt.serializer.generic.portMap import PortMap
from hwt.serializer.utils import maxStmId
from hwt.serializer.vhdl.keywords import VHLD_KEYWORDS
from hwt.serializer.vhdl.ops import VhdlSerializer_ops
from hwt.serializer.vhdl.statements import VhdlSerializer_statements
from hwt.serializer.vhdl.tmplContainer import VhdlTmplContainer
from hwt.serializer.vhdl.types import VhdlSerializer_types
from hwt.serializer.vhdl.utils import VhdlVersion
from hwt.serializer.vhdl.value import VhdlSerializer_Value
from hwt.synthesizer.param import getParam


class VhdlNameScope(NameScope):
    RE_MANY_UNDERSCORES = re.compile(r"(_{2,})")

    def checkedName(self, actualName, actualObj, isGlobal=False):
        actualName = self.RE_MANY_UNDERSCORES.sub(r"_", actualName)
        return NameScope.checkedName(self, actualName, actualObj,
                                     isGlobal=isGlobal)


class VhdlSerializer(VhdlTmplContainer, VhdlSerializer_Value,
                     VhdlSerializer_ops, VhdlSerializer_types,
                     VhdlSerializer_statements, GenericSerializer):
    VHDL_VER = VhdlVersion.v2002
    _keywords_dict = {kw: LangueKeyword() for kw in VHLD_KEYWORDS}
    fileExtension = '.vhd'

    @classmethod
    def getBaseNameScope(cls):
        s = VhdlNameScope(True)
        s.setLevel(1)
        s[0].update(cls._keywords_dict)
        return s

    @classmethod
    def Architecture(cls, arch, ctx):
        variables = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: (x.name, x._instId))
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        arch.components.sort(key=lambda x: x.name)
        arch.componentInstances.sort(key=lambda x: x._name)

        childCtx = ctx.withIndent()

        for v in arch.variables:
            t = v._dtype
            # if type requires extra definition
            if isinstance(t, (HEnum, HArray)) and t not in extraTypes:
                extraTypes.add(v._dtype)
                extraTypes_serialized.append(
                    cls.HdlType(t, childCtx, declaration=True))

            v.name = ctx.scope.checkedName(v.name, v)
            serializedVar = cls.SignalItem(v, childCtx, declaration=True)
            variables.append(serializedVar)

        for p in arch.processes:
            procs.append(cls.HWProcess(p, childCtx))

        # architecture names can be same for different entities
        # arch.name = scope.checkedName(arch.name, arch, isGlobal=True)

        uniqComponents = list(map(lambda x: x[1][0],
                                  groupedby(arch.components,
                                            lambda c: c.name)))
        uniqComponents.sort(key=lambda c: c.name)
        components = list(map(lambda c: cls.Component(c, childCtx),
                              uniqComponents))

        componentInstances = list(
            map(lambda c: cls.ComponentInstance(c, childCtx),
                arch.componentInstances))

        return cls.architectureTmpl.render(
            indent=getIndent(ctx.indent),
            entityName=arch.getEntityName(),
            name=arch.name,
            variables=variables,
            extraTypes=extraTypes_serialized,
            processes=procs,
            components=components,
            componentInstances=componentInstances
        )

    @classmethod
    def comment(cls, comentStr):
        return "--" + comentStr.replace("\n", "\n--")

    @classmethod
    def Component(cls, entity, ctx):
        return cls.componentTmpl.render(
            indent=getIndent(ctx.indent),
            ports=[cls.PortItem(pi, ctx) for pi in entity.ports],
            generics=[cls.GenericItem(g, ctx) for g in entity.generics],
            entity=entity
        )

    @classmethod
    def ComponentInstance(cls, entity, ctx):
        portMaps = []
        for pi in entity.ports:
            pm = PortMap.fromPortItem(pi)
            portMaps.append(pm)

        genericMaps = []
        for g in entity.generics:
            gm = MapExpr(g, g._val)
            genericMaps.append(gm)

        if len(portMaps) == 0:
            raise SerializerException("Incomplete component instance")

        entity._name = ctx.scope.checkedName(entity._name, entity)
        return cls.componentInstanceTmpl.render(
            indent=getIndent(ctx.indent),
            instanceName=entity._name,
            entity=entity,
            portMaps=[cls.PortConnection(x, ctx) for x in portMaps],
            genericMaps=[cls.MapExpr(x, ctx) for x in genericMaps]
        )

    @classmethod
    def Entity(cls, ent, ctx):
        generics, ports = cls.Entity_prepare(ent, ctx)

        entVhdl = cls.entityTmpl.render(
            indent=getIndent(ctx.indent),
            name=ent.name,
            ports=ports,
            generics=generics
        )

        doc = ent.__doc__
        if doc and id(doc) != id(Entity.__doc__):
            doc = cls.comment(doc) + "\n"
            return doc + entVhdl
        else:
            return entVhdl

    @classmethod
    def GenericItem(cls, g, ctx):
        s = "%s: %s" % (g.name, cls.HdlType(g._dtype, ctx))
        if g.defaultVal is None:
            return s
        else:
            return "%s := %s" % (
                s, cls.Value(getParam(g.defaultVal).staticEval(), ctx))

    @classmethod
    def PortConnection(cls, pc, ctx):
        if pc.portItem._dtype != pc.sig._dtype:
            raise SerializerException(
                "Port map %s is nod valid"
                " (types does not match)  (%r, %r)" % (
                    "%s => %s" % (pc.portItem.name, cls.asHdl(pc.sig, ctx)),
                    pc.portItem._dtype, pc.sig._dtype))
        return "%s => %s" % (pc.portItem.name, cls.asHdl(pc.sig, ctx))

    @classmethod
    def DIRECTION(cls, d):
        return d.name

    @classmethod
    def PortItem(cls, pi, ctx):
        return "%s: %s %s" % (pi.name, cls.DIRECTION(pi.direction),
                              cls.HdlType(pi._dtype, ctx))

    @classmethod
    def MapExpr(cls, m, ctx):
        return "%s => %s" % (m.compSig.name, cls.asHdl(m.value, ctx))



from hwt.serializer.ip_packager.exprSerializer import \
    VivadoTclExpressionSerializer
from hwt.serializer.ip_packager.helpers import appendSpiElem,\
    appendStrElements, findS, mkSpiElm, spi_ns_prefix, appendSpiArray
from hwt.serializer.ip_packager.otherXmlObjs import Value
import xml.etree.ElementTree as etree


class FileSetRef():
    @classmethod
    def fromElem(cls, elm):
        self = cls()
        self.localName = findS(elm, 'localName').text
        return self

    def asElem(self):
        e = etree.Element(spi_ns_prefix + "fileSetRef")
        appendSpiElem(e, "localName").text = self.localName
        return e


class View():
    _requiredVal = ["name", "displayName", "envIdentifier"]
    _optionalVal = ["language", "modelName"]

    @classmethod
    def fromElem(cls, elm):
        self = cls()
        for n in self._requiredVal:
            e = findS(elm, n)
            if e is None:
                raise Exception("View is missing " + n)
            setattr(self, n, e.text)
        for n in self._optionalVal:
            e = findS(elm, n)
            if e is None:
                continue
            setattr(self, n, e.text)

        self.fileSetRef = FileSetRef.fromElem(findS(elm, "fileSetRef"))
        return self

    def asElem(self):
        e = mkSpiElm("view")
        appendStrElements(e, self,
                          reqPropNames=self._requiredVal,
                          optPropNames=self._optionalVal)
        e.append(self.fileSetRef.asElem())
        return e


class ModelParameter():
    def __init__(self, name: str, displayName: str, datatype: str,
                 value: Value):
        self.name = name
        self.displayName = displayName
        self.datatype = datatype
        self.value = value

    @classmethod
    def fromGeneric(cls, g):
        val = Value.fromGeneric("MODELPARAM_VALUE.", g,
                                Value.RESOLVE_GENERATED)

        def createTmpVar(suggestedName, dtype):
            raise NotImplementedError(
                "Value of generic %s can not be converted do"
                "ipcore format (%r)" % (g.name, val))
        return cls(g.name,
                   g.name.replace("_", " "),
                   VivadoTclExpressionSerializer.HdlType(g._dtype, createTmpVar
                                                         ).lower(), val)

    def asElem(self):
        e = mkSpiElm("modelParameter")
        e.attrib["spirit:dataType"] = self.datatype
        appendStrElements(e, self,
                          reqPropNames=['name', "displayName"])
        e.append(self.value.asElem())
        return e


class Model():

    def __init__(self, vhdl_syn_fileSetName, vhdl_sim_fileSetName,
                 tcl_fileSetName):
        self.views = []
        self.ports = []
        self.modelParameters = []
        self.vhdl_syn_fileSetName = vhdl_syn_fileSetName
        self.vhdl_sim_fileSetName = vhdl_sim_fileSetName
        self.tcl_fileSetName = tcl_fileSetName

    def addDefaultViews(self, unit):
        viewsTemplate = ("""
       <views xmlns:xilinx="http://www.xilinx.com" xmlns:spirit="http://www.spiritconsortium.org/XMLSchema/SPIRIT/1685-2009" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
      <spirit:view>
        <spirit:name>xilinx_vhdlsynthesis</spirit:name>
        <spirit:displayName>VHDL Synthesis</spirit:displayName>
        <spirit:envIdentifier>vhdlSource:vivado.xilinx.com:synthesis</spirit:envIdentifier>
        <spirit:language>vhdl</spirit:language>
        <spirit:modelName>{0}</spirit:modelName>
        <spirit:fileSetRef>
          <spirit:localName>""" + self.vhdl_syn_fileSetName + """</spirit:localName>
        </spirit:fileSetRef>
      </spirit:view>
      <spirit:view>
        <spirit:name>xilinx_vhdlbehavioralsimulation</spirit:name>
        <spirit:displayName>VHDL Simulation</spirit:displayName>
        <spirit:envIdentifier>vhdlSource:vivado.xilinx.com:simulation</spirit:envIdentifier>
        <spirit:language>vhdl</spirit:language>
        <spirit:modelName>{0}</spirit:modelName>
        <spirit:fileSetRef>
          <spirit:localName>""" + self.vhdl_sim_fileSetName + """</spirit:localName>
        </spirit:fileSetRef>
      </spirit:view>
      <spirit:view>
        <spirit:name>xilinx_xpgui</spirit:name>
        <spirit:displayName>UI Layout</spirit:displayName>
        <spirit:envIdentifier>:vivado.xilinx.com:xgui.ui</spirit:envIdentifier>
        <spirit:fileSetRef>
          <spirit:localName>""" + self.tcl_fileSetName + """</spirit:localName>
        </spirit:fileSetRef>
      </spirit:view>
    </views>
        """).format(unit._name)
        for v in etree.fromstring(viewsTemplate):
            v = View.fromElem(v)
            v.modelName = unit._name
            self.views.append(v)
        for g in unit._entity.generics:
            mp = ModelParameter.fromGeneric(g)
            self.modelParameters.append(mp)

    # @classmethod
    # def fromElem(cls, elm):
    #    self = cls()
    #    views = findS(elm, "views")
    #    for vElm in views:
    #        self.views.append(View.fromElem(vElm))
    #    ports = findS(elm, "ports")
    #    for p in ports:
    #        self.ports.append(Port.fromElem(p))
    #    return self

    def asElem(self):
        e = mkSpiElm("model")
        appendSpiArray(e, 'views', self.views)
        appendSpiArray(e, 'ports', self.ports)
        appendSpiArray(e, 'modelParameters', self.modelParameters)

        return e



from hwt.hdl.types.defs import SLICE
from hwt.hdl.types.typeCast import toHVal
from hwt.simulator.simSignal import SimSignal
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class IndexSimSignalProxy(SimSignal):
    """
    Proxy which allows place indexing operations on signal
    """

    def __init__(self, name, baseSignal, dtype, upperIndex, lowerIndex=None):
        """
        :param lowerIndex: if this is none only upper index will be used
            like sig[upperIndex] else range select like
            sig[upperIndex:lowerIndex]
        """
        defaultVal = dtype.fromPy(None)

        self.name = name
        self._dtype = dtype
        self.defaultVal = defaultVal

        self.hidden = False
        self._writeCallbacks = []
        self._writeCallbacksToEn = []
        self.simSensProcs = set()
        self.simRisingSensProcs = set()
        self.simFallingSensProcs = set()
        self._signal = baseSignal
        if lowerIndex is None:
            self.__index = toHVal(upperIndex)
        else:
            self.__index = SLICE.fromPy([upperIndex, lowerIndex])
        self._setDefValue()

    def _generic_val_get(self, v):
        return v._getitem__val(self.__index)

    def _val_get(self):
        return self._generic_val_get(self._signal._val)

    def _oldVal_get(self):
        return self._generic_val_get(self._signal._oldVal)

    _val = property(_val_get)
    _oldVal = property(_oldVal_get)

    def simPropagateChanges(self, simulator):
        raise NotImplementedError(
            "Call this function on real signals not on this proxy")

    def simUpdateVal(self, simulator, valUpdater):
        """
        Method called by simulator to update new value for this object
        we are only delegating update on parent signal
        """
        dirtyFlag, newVal = valUpdater(
            self._generic_val_get(self._signal._oldVal))
        newVal.updateTime = simulator.now

        if dirtyFlag:
            # perform registration of new write callbacks
            if self._writeCallbacksToEn:
                SimSignal._loadWriteCallbacks(self)

            # run write callbacks we have to create new list to allow
            # registering of new call backs in callbacks
            for c in self._writeCallbacks:
                if c:
                    # simulation processes
                    simulator.process(c(simulator))

            def updateParent(oldVal):
                oldVal[self.__index] = newVal
                return True, oldVal

            self._signal.simUpdateVal(simulator, updateParent)

    def _setDefValue(self):
        v = self.defaultVal
        if isinstance(v, RtlSignalBase):
            v = v.staticEval()

        s = self._signal
        s._val[self.__index] = v.clone()
        s._oldVal[self.__index] = self._val.clone()
        s._oldVal.vldMask = 0

    def __repr__(self):
        return "<IndexSimSignalProxy %s[%r]>" % (
            self._signal.name, self.__index)





class SignalOps(object):
    """
    Operands for Signal interface,
    these operands are delegated on RtlSignal object for this interface
    """

    def _auto_cast(self, toT):
        return self._sig._auto_cast(toT)

    def _reinterpret_cast(self, toT):
        return self._sig._reinterpret_cast(toT)

    # events
    def _onRisingEdge(self):
        return self._sig._onRisingEdge()

    def _onFallingEdge(self):
        return self._sig._onFallingEdge()

    # comparison
    def _isOn(self):
        """
        convert this signal to hBool
        """
        return self._sig._isOn()

    def _eq(self, other):
        """
        Equality operator "==" is not used because it would damage python
        ecosystem
        """
        return self._sig._eq(other)

    def __ne__(self, other):
        """!="""
        return self._sig.__ne__(other)

    def __gt__(self, other):
        """>"""
        return self._sig.__gt__(other)

    def __lt__(self, other):
        """<"""
        return self._sig.__lt__(other)

    def __ge__(self, other):
        """>="""
        return self._sig.__ge__(other)

    def __le__(self, other):
        """<="""
        return self._sig.__le__(other)

    # bitwise
    def __invert__(self):
        """~ operator - logical negation for one bit signals and hBool
           bitwise inversion for wider signals """
        return self._sig.__invert__()

    def __and__(self, other):
        """
        & operator - logical 'and' for one bit signals and hBool
        bitwise and for wider signals
        """
        return self._sig.__and__(other)

    def __xor__(self, other):
        """
        ^ operator - logical '!=' for one bit signals and hBool
        bitwise and for wider signals
        """
        return self._sig.__xor__(other)

    def __or__(self, other):
        """
        | operator - logical 'or' for one bit signals and hBool
        bitwise and for wider signals
        """
        return self._sig.__or__(other)

    # arithmetic
    def __neg__(self):
        return self._sig.__neg__()

    def __add__(self, other):
        return self._sig.__add__(other)

    def __sub__(self, other):
        return self._sig.__sub__(other)

    def __mul__(self, other):
        return self._sig.__mul__(other)

    # hdl centric
    def _reversed(self):
        """
        reverse bitorder
        """
        return self._sig._reversed()

    def _concat(self, *others):
        """
        concatenate signals to one big one
        """
        return self._sig._concat(*others)

    def __getitem__(self, key):
        """
        [] operator key can be slice or index
        """
        return self._sig.__getitem__(key)

    def _ternary(self, ifTrue, ifFalse):
        return self._sig._ternary(ifTrue, ifFalse)

    def __call__(self, source):
        """
        connect this signal to driver

        :attention: it is not call of function it is operator of assignment
        :return: list of assignments
        """
        assert self._isAccessible
        return self._sig(source)



from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.typeShortcuts import hInt
from hwt.hdl.types.typeCast import toHVal
from hwt.serializer.generic.indent import getIndent


class VhdlSerializer_types():

    @classmethod
    def HdlType_bool(cls, typ, ctx, declaration=False):
        assert not declaration
        return "BOOLEAN"

    @classmethod
    def HdlType_bits(cls, typ, ctx, declaration=False):
        disableRange = False
        bitLength = typ.bit_length()
        w = typ.width
        isVector = typ.forceVector or bitLength > 1

        if typ.signed is None:
            if isVector:
                name = 'STD_LOGIC_VECTOR'
            else:
                return 'STD_LOGIC'
        elif typ.signed:
            name = "SIGNED"
        else:
            name = 'UNSIGNED'

        if disableRange:
            constr = ""
        elif isinstance(w, int):
            constr = "(%d DOWNTO 0)" % (w - 1)
        else:
            o = Operator(AllOps.SUB, (w, hInt(1)))
            constr = "(%s DOWNTO 0)" % cls.Operator(o, ctx)
        return name + constr

    @classmethod
    def HdlType_enum(cls, typ, ctx, declaration=False):
        buff = []
        if declaration:
            try:
                name = typ.name
            except AttributeError:
                name = "enumT_"
            typ.name = ctx.scope.checkedName(name, typ)

            buff.extend(["%sTYPE " % getIndent(ctx.indent),
                         typ.name.upper(), ' IS ('])
            # [TODO] check enum values names
            buff.append(", ".join(typ._allValues))
            buff.append(")")
            return "".join(buff)
        else:
            return typ.name

    @classmethod
    def HdlType_array(cls, typ, ctx, declaration=False):
        if declaration:
            try:
                name = typ.name
            except AttributeError:
                name = "arrT_"

            typ.name = ctx.scope.checkedName(name, typ)

            return "%sTYPE %s IS ARRAY ((%s) DOWNTO 0) OF %s" % \
                (getIndent(ctx.indent),
                 typ.name,
                 cls.asHdl(toHVal(typ.size) - 1, ctx),
                 cls.HdlType(typ.elmType, ctx, declaration=declaration))
        else:
            try:
                return typ.name
            except AttributeError:
                # [TODO]
                # sometimes we need to debug expression and we need temporary
                # type name this may be risk and this should be done
                # by extra debug serializer
                return "arrT_%d" % id(typ)

    @classmethod
    def HdlType_int(cls, typ, ctx, declaration=False):
        return "INTEGER"



from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.defs import BOOL, INT
from hwt.hdl.types.slice import Slice
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value
from hwt.synthesizer.param import evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class HArrayVal(Value):
    """
    Class for values of array HDL type
    """

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        :param val: None or dictionary {index:value} or iterrable of values
        :param vldMask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        size = evalParam(typeObj.size)
        if isinstance(size, Value):
            size = int(size)

        elements = {}
        if vldMask == 0:
            val = None

        if val is None:
            pass
        elif isinstance(val, dict):
            for k, v in val.items():
                if not isinstance(k, int):
                    k = int(k)
                elements[k] = typeObj.elmType.fromPy(v)
        else:
            for k, v in enumerate(val):
                if isinstance(v, RtlSignalBase):  # is signal
                    assert v._dtype == typeObj.elmType
                    e = v
                else:
                    e = typeObj.elmType.fromPy(v)
                elements[k] = e

        _mask = int(bool(val))
        if vldMask is None:
            vldMask = _mask
        else:
            assert (vldMask == _mask)

        return cls(elements, typeObj, vldMask)

    def __hash__(self):
        return hash((self._dtype, self.updateTime))
        # return hash((self._dtype, self.val, self.vldMask, self.updateTime))

    def _isFullVld(self):
        return self.vldMask == 1

    def _getitem__val(self, key):
        """
        :atention: this will clone item from array, iterate over .val
            if you need to modify items
        """
        try:
            kv = key.val
            if not key._isFullVld():
                raise KeyError()
            else:
                if kv >= self._dtype.size:
                    raise IndexError()

            return self.val[kv].clone()
        except KeyError:
            return self._dtype.elmType.fromPy(None)

    def __getitem__(self, key):
        iamVal = isinstance(self, Value)
        key = toHVal(key)
        isSLICE = isinstance(key, Slice.getValueCls())

        if isSLICE:
            raise NotImplementedError()
        elif isinstance(key, RtlSignalBase):
            key = key._auto_cast(INT)
        elif isinstance(key, Value):
            pass
        else:
            raise NotImplementedError(
                "Index operation not implemented for index %r" % (key))

        if iamVal and isinstance(key, Value):
            return self._getitem__val(key)

        return Operator.withRes(AllOps.INDEX, [self, key], self._dtype.elmType)

    def _setitem__val(self, index, value):
        self.updateTime = max(index.updateTime, value.updateTime)
        if index._isFullVld():
            self.val[index.val] = value.clone()
        else:
            self.val = {}

    def __setitem__(self, index, value):
        """
        Only syntax sugar for user, not used inside HWT

        * In HW design is not used (__getitem__ returns "reference"
            and it is used)

        * In simulator is used _setitem__val directly
        """
        if isinstance(index, int):
            index = INT.fromPy(index)
        else:
            assert isinstance(self, Value)
            assert index._dtype == INT, index._dtype

        if not isinstance(value, Value):
            value = self._dtype.elmType.fromPy(value)
        else:
            assert value._dtype == self._dtype.elmType, (
                value._dtype, self._dtype.elmType)

        return self._setitem__val(index, value)

    def __len__(self):
        return int(self._dtype.size)

    def _eq__val(self, other):
        assert self._dtype.elmType == other._dtype.elmType
        assert self._dtype.size == other._dtype.size

        eq = True
        vld = 1
        updateTime = -1
        keysA = set(self.val)
        keysB = set(other.val)
        sharedKeys = keysA.union(keysB)

        lsh = len(sharedKeys)
        if (lsh == int(self._dtype.size)
                and len(keysA) == lsh
                and len(keysB) == lsh):
            for k in sharedKeys:
                a = self.val[k]
                b = other.val[k]

                eq = eq and a == b
                if not eq:
                    break
                vld = vld & a.vldMask & b.vldMask
                updateTime = max(updateTime, a.updateTime, b.updateTime)
        else:
            eq = False
            vld = 0

        return BOOL.getValueCls()(eq, BOOL, vld, updateTime)

    def _eq(self, other):
        assert isinstance(other, HArrayVal)
        return self._eq__val(other)



from datetime import datetime
import sys

from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bool import HBool
from hwt.hdl.types.enum import HEnum
from hwt.simulator.hdlSimConfig import HdlSimConfig
from hwt.simulator.vcdWritter import VcdWritter


class VcdHdlSimConfig(HdlSimConfig):
    supported_type_classes = (HBool, Bits, HEnum)

    def __init__(self, dumpFile=sys.stdout):
        self.vcdWritter = VcdWritter(dumpFile)
        self.logPropagation = False
        self.logApplyingValues = False

    def vcdRegisterUnit(self, unit):
        with self.vcdWritter.module(unit._name) as m:
            for se in unit._cntx.signals:
                if isinstance(se._dtype, self.supported_type_classes):
                    m.var(se)

            for u in unit._units:
                self.vcdRegisterUnit(u)

    def beforeSim(self, simulator, synthesisedUnit):
        """
        This method is called before first step of simulation.
        """
        self.vcdWritter.date(datetime.now())
        self.vcdWritter.timescale(1)

        self.vcdRegisterUnit(synthesisedUnit)
        self.vcdWritter.enddefinitions()

    def logChange(self, nowTime, sig, nextVal):
        """
        This method is called for every value change of any signal.
        """
        try:
            self.vcdWritter.change(nowTime, sig, nextVal)
        except KeyError:
            # not every signal has to be registered
            pass



from hwt.simulator.shortcuts import OnRisingCallbackLoop
from hwt.synthesizer.exceptions import IntfLvlConfErr


class AgentBase():
    """
    Base class of agent of interface like in UVM
    driver is used for slave interfaces
    monitor is used for master interfaces

    :ivar intf: interface assigned to this agent
    :ivar enable: flag to enable/disable this agent
    """

    def __init__(self, intf):
        self.intf = intf
        self._enabled = True
        self._debugOutput = None

    def setEnable(self, en, sim):
        self._enabled = en

    def getEnable(self):
        return self._enabled

    def _debug(self, out):
        self._debugOutput = out

    def getDrivers(self):
        """
        Called before simulation to collect all drivers of interfaces
        from this agent
        """
        return [self.driver]

    def getMonitors(self):
        """
        Called before simulation to collect all monitors of interfaces
        from this agent
        """
        return [self.monitor]

    def driver(self, s):
        """
        Implement this method to drive your interface
        in simulation/verification
        """
        raise NotImplementedError()

    def monitor(self, s):
        """
        Implement this method to monitor your interface
        in simulation/verification
        """
        raise NotImplementedError()


class AgentWitReset(AgentBase):
    def __init__(self, intf, allowNoReset=False):
        super().__init__(intf)
        self._discoverReset(allowNoReset)

    def _discoverReset(self, allowNoReset):
        try:
            rst = self.intf._getAssociatedRst()
            self.rst = rst._sigInside
            self.rstOffIn = int(rst._dtype.negated)
            self.notReset = self._notReset
        except IntfLvlConfErr:
            self.rst = None
            self.notReset = self._notReset_dummy

            if allowNoReset:
                pass
            else:
                raise

    def _notReset_dummy(self, sim):
        return True

    def _notReset(self, sim):
        rstVal = sim.read(self.rst).val
        return rstVal == self.rstOffIn


class SyncAgentBase(AgentWitReset):
    """
    Agent which discovers clk, rst signal and runs only
    at specified edge of clk

    :attention: requires clk and rst/rstn signal
        (if you do not have any create simulation wrapper with it)
    """
    SELECTE_EDGE_CALLBACK = OnRisingCallbackLoop

    def __init__(self, intf, allowNoReset=False):
        super().__init__(intf, allowNoReset=allowNoReset)

        # resolve clk and rstn
        self.clk = self.intf._getAssociatedClk()._sigInside

        # run monitor, driver only on rising edge of clk
        c = self.SELECTE_EDGE_CALLBACK
        self.monitor = c(self.clk, self.monitor, self.getEnable)
        self.driver = c(self.clk, self.driver, self.getEnable)

    def setEnable_asDriver(self, en, sim):
        self._enabled = en
        self.driver.setEnable(en, sim)

    def setEnable_asMonitor(self, en, sim):
        self._enabled = en
        self.monitor.setEnable(en, sim)

    def getDrivers(self):
        self.setEnable = self.setEnable_asDriver
        return AgentBase.getDrivers(self)

    def getMonitors(self):
        self.setEnable = self.setEnable_asMonitor
        return AgentBase.getMonitors(self)




class ResourceError(Exception):
    """
    Resource of this kind does not exists in current hardware
    """


class Unconnected():
    pass


class ResourceMUX():
    def __init__(self, bitWidth, inputs):
        self.bitWidth = bitWidth
        self.inputs = inputs


class ResourceFF():
    pass


class ResourceFFwithMux():
    pass


class ResourceLatch():
    pass


class ResourceLatchWithMux():
    pass


class ResourceAsyncRAM():
    pass


class ResourceAsyncROM():
    pass


class ResourceRAM():
    """
    Specifier of type of RAM like memory
    """

    def __init__(self, width, items,
                 rwSyncPorts, rSyncPorts, wSyncPorts,
                 rwAsyncPorts, rAsyncPorts, wAsyncPorts):
        self.width = width
        self.items = items
        self.rwSyncPorts = rwSyncPorts
        self.rSyncPorts = rSyncPorts
        self.wSyncPorts = wSyncPorts
        self.rwAsyncPorts = rwAsyncPorts
        self.rAsyncPorts = rAsyncPorts
        self.wAsyncPorts = wAsyncPorts

    def __hash__(self):
        return hash((self.width, self.items,
                     self.rwSyncPorts, self.rSyncPorts,
                     self.wSyncPorts,
                     self.rwAsyncPorts, self.rAsyncPorts,
                     self.wAsyncPorts))

    def __eq__(self, other):
        return (
            self.width == other.width and
            self.items == other.items and
            self.rwSyncPorts == other.rwSyncPorts and
            self.rSyncPorts == other.rSyncPorts and
            self.wSyncPorts == other.wSyncPorts and
            self.rwAsyncPorts == other.rwAsyncPorts and
            self.rAsyncPorts == other.rAsyncPorts and
            self.wAsyncPorts == other.wAsyncPorts
        )

    def __repr__(self):
        return ("<%s, %dbx%d, syncPorts: (rw:%d, r:%d, w:%d), asyncPorts:"
                " (rw:%d, r:%d, w:%d)>") % (
            self.__class__.__name__, self.width, self.items,
            self.rwSyncPorts, self.rSyncPorts, self.wSyncPorts,
            self.rwAsyncPorts, self.rAsyncPorts, self.wAsyncPorts
        )


class ResourceROM():
    pass




from typing import Optional, Any

from hwt.hdl.types.defs import INT, STR, BOOL
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.value import Value
from hwt.hdl.variables import SignalItem
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase


defaultPyConversions = {int: INT,
                        str: STR,
                        bool: BOOL}


def toHVal(op: Any, suggestedType: Optional[HdlType]=None):
    """Convert python or hdl value/signal object to hdl value/signal object"""
    if isinstance(op, Value) or isinstance(op, SignalItem):
        return op
    elif isinstance(op, InterfaceBase):
        return op._sig
    else:
        if isinstance(op, int):
            if suggestedType is not None:
                return suggestedType.fromPy(op)

            if op >= 1 << 31:
                raise TypeError(
                    "Number %d is too big to fit in 32 bit integer of HDL"
                    " use Bits type instead" % op)
            elif op < -(1 << 31):
                raise TypeError(
                    "Number %d is too small to fit in 32 bit integer"
                    " of HDL use Bits type instead" % op)
        try:
            hType = defaultPyConversions[type(op)]
        except KeyError:
            hType = None

        if hType is None:
            raise TypeError("%s" % (op.__class__))

        return hType.fromPy(op)



from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.structValBase import StructValBase
from hwt.serializer.generic.indent import getIndent


class HStructFieldMeta():
    def __init__(self, split=False):
        self.split = split

    def __eq__(self, other):
        return self.split == other.split

    def __hash__(self):
        return hash(self.split)


class HStructField(object):
    def __init__(self, typ, name, meta=None):
        assert isinstance(name, str) or name is None, name
        assert isinstance(typ, HdlType), typ
        self.name = name
        self.dtype = typ
        self.meta = meta

    def __repr__(self):
        return "<HStructField %r, %s>" % (self.dtype, self.name)


protectedNames = {"clone", "staticEval", "fromPy", "_dtype"}


class HStruct(HdlType):
    """
    HDL structure type

    :ivar fields: tuple of HStructField instances in this struct
    :ivar name: name of this HStruct type
    :ivar valueCls: Class of value for this type as usual
        in HdlType implementations
    """
    def __init__(self, *template, name=None):
        """
        :param template: list of tuples (type, name) or HStructField objects
            name can be None (= padding)
        :param name: optional name used for debugging purposes
        """
        fields = []
        self.name = name
        fieldNames = []
        bit_length = 0
        for f in template:
            try:
                field = HStructField(*f)
            except TypeError:
                field = f
            if not isinstance(field, HStructField):
                raise TypeError("Template for struct field %s is not in valid format" % repr(f))

            fields.append(field)
            if field.name is not None:
                fieldNames.append(field.name)

            t = field.dtype
            if bit_length is not None:
                try:
                    _bit_length = t.bit_length()
                    bit_length += _bit_length
                except TypeError:
                    bit_length = None

        self.fields = tuple(fields)
        self.__bit_length_val = bit_length

        usedNames = set(fieldNames)
        assert not protectedNames.intersection(usedNames),\
            protectedNames.intersection(usedNames)

        class StructVal(StructValBase):
            __slots__ = fieldNames

        if name is not None:
            StructVal.__name__ = name + "Val"

        self.valueCls = StructVal

    def bit_length(self):
        bl = self.__bit_length_val
        if bl is None:
            raise TypeError("Can not request bit_lenght on size"
                            " which has not fixed size")
        else:
            return self.__bit_length_val

    def getValueCls(self):
        return self.valueCls

    @classmethod
    def get_reinterpret_cast_fn(cls):
        from hwt.hdl.types.structCast import hstruct_reinterpret
        return hstruct_reinterpret

    def __fields__eq__(self, other):
        if len(self.fields) != len(other.fields):
            return False
        for sf, of in zip(self.fields, other.fields):
            if (sf.name != of.name or
                    sf.dtype != of.dtype or
                    sf.meta != of.meta):
                return False
        return True

    def __eq__(self, other):
        return self is other or (
            type(self) is type(other) and
            self.bit_length() == other.bit_length() and
            self.__fields__eq__(other))

    def __hash__(self):
        return hash(id(self))

    def __add__(self, other):
        """
        override of addition, merge struct into one
        """
        assert isinstance(other, HStruct)
        return HStruct(*self.fields, *other.fields)

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not None is used as a additional
            information about on which address this type is stored
            (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and HArray)
        """
        if self.name:
            name = self.name + " "
        else:
            name = ""

        myIndent = getIndent(indent)
        childIndent = getIndent(indent + 1)
        header = "%sstruct %s{" % (myIndent, name)

        buff = [header, ]
        for f in self.fields:
            if withAddr is not None:
                addrTag = " // start:0x%x(bit) 0x%x(byte)" % (
                            withAddr, withAddr // 8)
            else:
                addrTag = ""

            if f.name is None:
                buff.append("%s//%r empty space%s" % (
                            childIndent, f.dtype, addrTag))
            else:
                buff.append("%s %s%s" % (
                               f.dtype.__repr__(indent=indent + 1,
                                                withAddr=withAddr,
                                                expandStructs=expandStructs),
                            f.name, addrTag))
            if withAddr is not None:
                withAddr += f.dtype.bit_length()

        buff.append("%s}" % (myIndent))
        return "\n".join(buff)




from functools import wraps
import sys

from hwt.hdl.types.defs import BIT
from hwt.hdl.types.enum import HEnum
from hwt.serializer.vhdl.serializer import VhdlSerializer


def dumpMethod(func):
    """decorator which takes functions return and write it as line to dumpFile
    """
    @wraps(func)
    def wrapped(*args, **kwrds):
        s = func(*args, **kwrds)
        if s is not None:
            self = args[0]
            self.dumpFile.write(s)
            self.dumpFile.write('\n')

    return wrapped


class VcdVarInfo():
    """Info about signal registered in vcd"""

    def __init__(self, _id, dtype):
        if isinstance(dtype, HEnum):
            self.width = 1
        else:
            self.width = dtype.bit_length()
        self.id = _id
        self._dtype = dtype


class VcdVarContext(dict):
    """Map of signals registered in this unit"""

    def __init__(self):
        super().__init__()
        self.nextId = 0
        self.idChars = [chr(i) for i in range(ord("!"), ord("~") + 1)]
        self.idCharsCnt = len(self.idChars)

    def idToStr(self, x):
        if x < 0:
            sign = -1
        elif x == 0:
            return self.idChars[0]
        else:
            sign = 1
        x *= sign
        digits = []
        while x:
            digits.append(self.idChars[x % self.idCharsCnt])
            x //= self.idCharsCnt
        if sign < 0:
            digits.append('-')
        digits.reverse()
        return ''.join(digits)

    def register(self, var):
        var_id = self.idToStr(self.nextId)
        if var in self:
            raise KeyError("%s is already registered" % (repr(var)))
        vInf = VcdVarInfo(var_id, var._dtype)
        self[var] = vInf
        self.nextId += 1
        return vInf


class VcdModule():
    """Vcd module - container for variables"""

    def __init__(self, dumpFile, _vars, name):
        self.name = name
        self.dumpFile = dumpFile
        self.vars = _vars

    def __enter__(self):
        self.header()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.footer()

    @dumpMethod
    def header(self):
        return "$scope module %s $end" % self.name

    @dumpMethod
    def var(self, sig):
        vInf = self.vars.register(sig)
        if isinstance(vInf._dtype, HEnum):
            sigType = 'real'
        else:
            sigType = 'wire'

        return "$var %s %d %s %s $end" % (sigType, vInf.width,
                                          vInf.id, sig.name)

    @dumpMethod
    def footer(self):
        return "$upscope $end"


class VcdWritter():
    def __init__(self, dumpFile=sys.stdout):
        self.dumpFile = dumpFile
        self.vars = VcdVarContext()
        self.lastTime = -1
        self.defaultTop = None

    @dumpMethod
    def date(self, text):
        return "$date\n   %s\n$end" % text

    @dumpMethod
    def version(self, text):
        return "$version   \n%s\n$end" % text

    @dumpMethod
    def timescale(self, picoSeconds):
        return "$timescale %dps $end" % picoSeconds

    def module(self, name):
        return VcdModule(self.dumpFile, self.vars, name)

    @dumpMethod
    def enddefinitions(self):
        return "$enddefinitions $end"

    @dumpMethod
    def setTime(self, t):
        lt = self.lastTime
        if lt == t:
            return
        elif lt < t:
            self.lastTime = t
            return "#%d" % (t)
        else:
            raise Exception("VcdWritter invalid time update %d -> %d" % (
                            lt, t))

    @dumpMethod
    def change(self, time, sig, newVal):
        self.setTime(time)
        varInfo = self.vars[sig]
        if isinstance(sig._dtype, HEnum):
            if newVal.vldMask:
                val = newVal.val
            else:
                val = "XXXX"
            frmt = "s%s %s"
        else:
            val = VhdlSerializer.BitString_binary(
                newVal.val, varInfo.width, newVal.vldMask)
            val = val.replace('"', "")

            if varInfo._dtype == BIT:
                frmt = "%s%s"
            else:
                frmt = "b%s %s"

        return frmt % (val, varInfo.id)




class TypeConversionErr(TypeError):
    pass


class ConfErr(Exception):
    pass


class IntfLvlConfErr(ConfErr):
    """
    Interface level synthesizer user configuration error
    """
    pass


class SigLvlConfErr(ConfErr):
    """
    Signal level synthesizer user configuration error
    """
    pass



from typing import Dict, Tuple

from hwt.hdl.assignment import Assignment
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.pyUtils.arrayQuery import arr_any
from hwt.serializer.resourceAnalyzer.resourceTypes import Unconnected, \
    ResourceFF, ResourceMUX, ResourceFFwithMux, ResourceLatch, ResourceRAM, \
    ResourceROM, ResourceLatchWithMux, ResourceAsyncRAM, ResourceAsyncROM
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


# tables for resolving of resource type updates
# sequential statements merging
resourceTransitions_override = {
    (Assignment, Assignment): Assignment,  # overwrite
    (Unconnected, Unconnected): Unconnected,

    (Assignment, Unconnected): Assignment,  # written in another command
    (Unconnected, Assignment): Assignment,

    (Assignment, ResourceMUX): ResourceMUX,  # written in another command
    (ResourceMUX, Assignment): ResourceMUX,


    (ResourceFF, Unconnected): ResourceFF,  # written in another command
    (Unconnected, ResourceFF): ResourceFF,

    (ResourceMUX, Unconnected): ResourceMUX,  # written in another command
    (Unconnected, ResourceMUX): ResourceMUX,

    # written in another command
    (ResourceFFwithMux, Assignment): ResourceFFwithMux,
    (Assignment, ResourceFFwithMux): ResourceFFwithMux,

    # written in another command
    (ResourceFFwithMux, Unconnected): ResourceFFwithMux,
    (Unconnected, ResourceFFwithMux): ResourceFFwithMux,


    (ResourceMUX, ResourceLatch): ResourceMUX,  # written in another command
    (ResourceLatch, ResourceMUX): ResourceMUX,

    (ResourceRAM, Unconnected): ResourceRAM,
    (Unconnected, ResourceRAM): ResourceRAM,

    (ResourceROM, Unconnected): ResourceROM,
    (Unconnected, ResourceROM): ResourceROM,

    (Unconnected, ResourceLatchWithMux): ResourceLatchWithMux,
    (Unconnected, ResourceLatch): ResourceLatchWithMux,

    (Assignment, ResourceLatchWithMux): ResourceMUX,

}


resourceTransitions_sameBranchLevel = {
    # current : now discovered
    (Assignment, Assignment): ResourceMUX,

    (Unconnected, Unconnected): Unconnected,
    (Assignment, Unconnected): ResourceLatch,
    (Unconnected, Assignment): ResourceLatch,

    (Assignment, ResourceMUX): ResourceMUX,
    (ResourceMUX, Assignment): ResourceMUX,

    (Assignment, Unconnected): ResourceLatchWithMux,
    (Unconnected, Assignment): ResourceLatchWithMux,

    (Assignment, ResourceLatch): ResourceLatchWithMux,
    (ResourceLatch, Assignment): ResourceLatchWithMux,

    (Assignment, ResourceLatchWithMux): ResourceLatchWithMux,
    (ResourceLatchWithMux, Assignment): ResourceLatchWithMux,

    (Unconnected, ResourceLatchWithMux): ResourceLatchWithMux,
    (ResourceLatchWithMux, Unconnected): ResourceLatchWithMux,

    (ResourceFF, ResourceFF): ResourceFFwithMux,
    (ResourceFF, Unconnected): ResourceFF,

    (Assignment, ResourceFF): ResourceFFwithMux,
    (ResourceFF, Assignment): ResourceFFwithMux,

    (ResourceMUX, ResourceFF): ResourceFFwithMux,
    (ResourceFF, ResourceMUX): ResourceFFwithMux,

    (ResourceMUX, Unconnected): ResourceLatchWithMux,
    (Unconnected, ResourceMUX): ResourceLatchWithMux,

    (ResourceFFwithMux, Unconnected): ResourceFFwithMux,
    (Unconnected, ResourceFFwithMux): ResourceFFwithMux,

    (ResourceRAM, Unconnected): ResourceRAM,
    (ResourceROM, Unconnected): ResourceROM,
    (ResourceAsyncRAM, Unconnected): ResourceAsyncRAM,
    (ResourceAsyncROM, Unconnected): ResourceAsyncROM,


}

resourceTransitions_memoryAnotherInput = {
    ResourceAsyncROM: ResourceAsyncRAM,
    ResourceAsyncRAM: ResourceAsyncRAM,
    ResourceROM: ResourceRAM,
    ResourceRAM: ResourceRAM,
}


def updateGuesFromAssignment(gues: Dict, assignment: Assignment) -> None:
    """
    Apply next assignment to this signal on current resoruce gues
    for this signal
    """
    sig = assignment.dst
    if assignment.indexes:
        isRam = arr_any(assignment.indexes,
                        lambda x: isinstance(x, RtlSignal) and not x._const)
    else:
        isRam = False

    try:
        current = gues[sig]
    except KeyError:
        if isRam:
            if assignment.isEventDependent:
                g = ResourceRAM
            else:
                g = ResourceAsyncRAM
        else:
            if assignment.isEventDependent:
                g = ResourceFF
            else:
                g = Assignment
        gues[sig] = g
        return

    if isRam:
        # [TODO] check for event dependency
        g = resourceTransitions_memoryAnotherInput[current]
    else:
        if assignment.isEventDependent:
            g = ResourceFF
        else:
            g = Assignment

        g = resourceTransitions_sameBranchLevel[(current, g)]
    gues[sig] = g


def mergeGues(gues: Dict, otherGues: Dict) -> None:
    """
    Merge resource gues dicts, "otherGues" into "gues"
    """
    update = {}
    for sig, g0 in gues.items():
        g1 = otherGues[sig]
        nextG = resourceTransitions_sameBranchLevel[(g0, g1)]
        if g0 is not nextG:
            update[sig] = nextG

    gues.update(update)


conversions = {AllOps.BitsAsSigned,
               AllOps.BitsAsUnsigned,
               AllOps.BitsAsVec,
               AllOps.BitsToInt,
               AllOps.IntToBits}


operatorsWithoutResource = conversions.union(
    {AllOps.CONCAT,
     AllOps.CALL,
     AllOps.FALLIGN_EDGE,
     AllOps.RISING_EDGE})


RamOrRomResources = {ResourceAsyncRAM,
                     ResourceAsyncROM,
                     ResourceRAM,
                     ResourceROM}


def findAssingmentOf(sig: RtlSignal) -> Assignment:
    """
    Walk endpoints to neares assignment
    (undirect/direct driver of this signal)
    """
    for ep in sig.endpoints:
        if isinstance(ep, Assignment):
            return ep
        elif isinstance(ep, Operator):
            return findAssingmentOf(ep.result)
        else:
            raise NotImplementedError()


class ResourceContext():
    """
    Container of informations about resources used in architecture

    :ivar unit: optional unit for which is this context build
    :ivar seen: set of seen objects
    :ivar resources: dictionary {type of resource: cnt}
    :ivar discoveredRamSignals: set of signals which seems to be some kind
        of RAM/ROM memory
    """

    def __init__(self, unit):
        self.unit = unit
        self.seen = set()
        self.resources = {}
        self.discoveredRamSignals = set()

    def registerOperator(self, op: Operator):
        w = op.operands[0]._dtype.bit_length()
        res = self.resources
        k = (op.operator, w)
        r = res.get(k, 0)
        res[k] = r + 1

    def registerMUX_known(self, width, inputsCnt):
        """
        Register MUX of known properties
        """
        k = (ResourceMUX, width, inputsCnt)
        muxs = self.resources.get(k, 0)
        self.resources[k] = muxs + 1

    def registerMUX(self, mux):
        """
        mux record is in format (self.MUX, n, m)
        where n is number of bits of this mux
        and m is number of possible inputs
        """
        inputs = len(mux.drivers)
        if inputs > 1:
            w = mux._dtype.bit_length()
            k = (ResourceMUX, w, inputs)
            muxs = self.resources.get(k, 0)
            self.resources[k] = muxs + 1

    def registerFF(self, ff):
        res = self.resources
        w = ff._dtype.bit_length()
        ffs = res.get(ResourceFF, 0)
        res[ResourceFF] = ffs + w

    def registerLatch(self, latch):
        res = self.resources
        w = latch._dtype.bit_length()
        latches = res.get(ResourceLatch, 0)
        res[ResourceLatch] = latches + w

    def _extractRamPorts(self, mem: RtlSignal) -> Dict[RtlSignal,
                                                       Tuple[int, int,
                                                             int, int]]:
        """
        Resolve address signals and read/write ports of memory
        :return: dict  address signal : [syncReads, syncWrites,
            asyncReads, asyncWrites]
        """

        addressSignals = {}
        # collect write ports
        for d in mem.drivers:
            assert isinstance(d, Assignment)
            assert len(d.indexes) == 1
            index = d.indexes[0]
            try:
                ports = addressSignals[index]
            except KeyError:
                ports = [0, 0, 0, 0]
                addressSignals[index] = ports

            if d.isEventDependent:
                ports[1] += 1
            else:
                ports[3] += 1
        # collect read ports
        for e in mem.endpoints:
            if isinstance(e, Assignment):
                assert len(e.indexes) == 1
                index = e.indexes
                isEventDependent = e.isEventDependent
            elif isinstance(e, Operator) and e.operator == AllOps.INDEX:
                index = e.operands[1]
                a = findAssingmentOf(e.result)
                isEventDependent = a.isEventDependent
            else:
                raise NotImplementedError(e)

            try:
                ports = addressSignals[index]
            except KeyError:
                ports = [0, 0, 0, 0]
                addressSignals[index] = ports

            # readportscnt++
            if isEventDependent:
                ports[0] += 1
            else:
                ports[2] += 1

        return addressSignals

    def registerMem(self, mem):
        # [TODO] find clks
        # walk all endpoints and drivers
        # for each address signal collect number of read/write pors
        # merge maximum or r/w ports to rw port for each address signal
        width = mem._dtype.elmType.bit_length()
        items = int(mem._dtype.size)

        # resolve conts of r/w ports
        rwSyncPorts = 0
        rSyncPorts = 0
        wSyncPorts = 0

        rwAsyncPorts = 0
        rAsyncPorts = 0
        wAsyncPorts = 0

        addressSignals = self._extractRamPorts(mem)
        for _, (r, w, asyncR, asyncW) in addressSignals.items():
            rw = min(r, w)
            rwSyncPorts += rw
            rSyncPorts += r - rw
            wSyncPorts += w - rw

            rw = min(asyncR, asyncW)
            rwAsyncPorts += rw
            rAsyncPorts += asyncR - rw
            wAsyncPorts += asyncW - rw

        # register ram resource for this mem
        m = ResourceRAM(width, items,
                        rwSyncPorts, rSyncPorts, wSyncPorts,
                        rwAsyncPorts, rAsyncPorts, wAsyncPorts)
        cnt = self.resources.get(m, 0)
        self.resources[m] = cnt + 1

        # if has sync reads merge FFs into this resource
        FFsInRam = (rwSyncPorts + rSyncPorts) * width
        if FFsInRam:
            ffs = self.resources[ResourceFF]
            if ffs == FFsInRam:
                del self.resources[ResourceFF]
            elif ffs > FFsInRam:
                self.resources[ResourceFF] = ffs - FFsInRam
            else:
                raise Exception(
                    "Incompatible ram description (read port did not found FFs"
                    " as expected)")

    def register(self, sig: RtlSignal, resourceGues):
        """
        Register resource(s) for the signal
        """
        if resourceGues is Assignment:
            pass
        elif resourceGues is ResourceMUX:
            self.registerMUX(sig)
        elif resourceGues is ResourceFF:
            self.registerFF(sig)
        elif resourceGues is ResourceFFwithMux:
            self.registerFF(sig)
            self.registerMUX(sig)
        elif resourceGues is ResourceLatch:
            self.registerLatch(sig)
        elif resourceGues is ResourceLatchWithMux:
            self.registerLatch(sig)
            self.registerMUX(sig)
        elif resourceGues in RamOrRomResources:
            self.discoveredRamSignals.add(sig)
        else:
            raise NotImplementedError(resourceGues)

    def finalize(self):
        """
        Resolve ports of discovered memories
        """
        for sig in self.discoveredRamSignals:
            self.registerMem(sig)
        self.gues = set()



from jinja2.environment import Environment
from jinja2.loaders import PackageLoader


class VerilogTmplContainer():
    tmplEnv = Environment(loader=PackageLoader(
        'hwt', 'serializer/verilog/templates'))

    moduleHeadTmpl = tmplEnv.get_template('module_head.v')
    moduleBodyTmpl = tmplEnv.get_template('module_body.v')
    processTmpl = tmplEnv.get_template('process.v')
    ifTmpl = tmplEnv.get_template("if.v")
    componentInstanceTmpl = tmplEnv.get_template("component_instance.v")
    switchTmpl = tmplEnv.get_template("switch.v")



from hwt.hdl.constants import DIRECTION
from hwt.hdl.types.bits import Bits
from hwt.interfaces.agents.tristate import TristateClkAgent, TristateAgent
from hwt.interfaces.std import Signal, Clk
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.param import Param


class TristateSig(Interface):
    """
    Tristate interface
    in order to make this a vector[0] instead of single bit
    use forceVector=True
    """

    def _config(self):
        self.DATA_WIDTH = Param(1)
        self.forceVector = False

    def _declr(self):
        t = Bits(self.DATA_WIDTH, self.forceVector)

        # connect
        self.t = Signal(dtype=t)
        # input
        self.i = Signal(dtype=t, masterDir=DIRECTION.IN)
        # output
        self.o = Signal(dtype=t)

    def _initSimAgent(self):
        self._ag = TristateAgent(self)


class TristateClk(Clk, TristateSig):
    def _getIpCoreIntfClass(self):
        raise NotImplementedError()

    def _initSimAgent(self):
        self._ag = TristateClkAgent(self)



from hwt.hdl.types.bool import HBool
from hwt.hdl.types.integer import Integer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.string import String
from hwt.hdl.types.slice import Slice

"""
Definitions of most common types
"""

BOOL = HBool()
INT = Integer()
BIT = Bits(width=1)
BIT_N = Bits(width=1, negated=True)
STR = String()
SLICE = Slice()



from copy import copy

from hwt.serializer.verilog.context import VerilogSerializerCtx


class SystemCCtx(VerilogSerializerCtx):
    """
    Context of SystemC serializer

    :ivar isTarget: specifies if current expression is part of target
        (= do not add .read())
    :ivar isSensitivityList: flag that specifies if serializer is
        in sensitivity list scope and signals are allowed
        to use event operators
    """

    def __init__(self, *args, **kwargs):
        super(SystemCCtx, self).__init__(*args, **kwargs)
        self.isTarget = False
        self.isSensitivityList = False

    def forTarget(self):
        c = copy(self)
        c.isTarget = True
        return c

    def forSensitivityList(self):
        c = copy(self)
        c.isTarget = True
        c.isSensitivityList = True
        return c



from hwt.hdl.assignment import Assignment
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.sliceUtils import slice_to_SLICE
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value
from hwt.synthesizer.exceptions import TypeConversionErr
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from typing import List


def tv(signal):
    """
    Value class for type of signal
    """
    return signal._dtype.getValueCls()


class RtlSignalOps():
    """
    Definitions of operators and other operator functions for RtlSignal

    :ivar _usedOps: cache for expressions with this signal
    """

    def _auto_cast(self, toT):
        return self._dtype.auto_cast(self, toT)

    def _reinterpret_cast(self, toT):
        return self._dtype.reinterpret_cast(self, toT)

    def naryOp(self, operator, opCreateDelegate, *otherOps):
        """
        Try lookup operator with this parameters in _usedOps
        if not found create new one and soter it in _usedOps
        """
        k = (operator, *otherOps)
        try:
            return self._usedOps[k]
        except KeyError:
            pass
        o = opCreateDelegate(self, *otherOps)
        self._usedOps[k] = o
        return o

    def __invert__(self):
        return self.naryOp(AllOps.NOT, tv(self).__invert__)

    def _onRisingEdge(self, now=None):
        return self.naryOp(AllOps.RISING_EDGE, tv(self)._onRisingEdge, now)

    def _onFallingEdge(self, now=None):
        return self.naryOp(AllOps.FALLIGN_EDGE, tv(self)._onFallingEdge, now)

    def _isOn(self):
        return self._auto_cast(BOOL)

    # conversions
    def _convSign(self, signed):
        return tv(self)._convSign(self, signed)

    def _signed(self):
        return tv(self)._signed(self)

    def _unsigned(self):
        return tv(self)._unsigned(self)

    def _vec(self):
        return tv(self)._vec(self)

    # logic
    def __and__(self, other):
        return self.naryOp(AllOps.AND, tv(self).__and__, other)

    def __xor__(self, other):
        return self.naryOp(AllOps.XOR, tv(self).__xor__, other)

    def __or__(self, other):
        return self.naryOp(AllOps.OR, tv(self).__or__, other)

    # cmp
    def _eq(self, other):
        """__eq__ is not overloaded because it will destroy hashability of object"""
        return self.naryOp(AllOps.EQ, tv(self)._eq, other)

    def __ne__(self, other):
        return self.naryOp(AllOps.NEQ, tv(self).__ne__, other)

    def __ge__(self, other):
        return self.naryOp(AllOps.GE, tv(self).__ge__, other)

    def __gt__(self, other):
        return self.naryOp(AllOps.GT, tv(self).__gt__, other)

    def __lt__(self, other):
        return self.naryOp(AllOps.LT, tv(self).__lt__, other)

    def __le__(self, other):
        return self.naryOp(AllOps.LE, tv(self).__le__, other)

    # arithmetic
    def __add__(self, other):
        return self.naryOp(AllOps.ADD, tv(self).__add__, other)

    def __sub__(self, other):
        return self.naryOp(AllOps.SUB, tv(self).__sub__, other)

    def __mul__(self, other):
        return self.naryOp(AllOps.MUL, tv(self).__mul__, other)

    def __floordiv__(self, divider):
        return self.naryOp(AllOps.DIV, tv(self).__floordiv__, divider)

    # selections
    def _downto(self, to):
        return self.naryOp(AllOps.DOWNTO, tv(self)._downto, to)

    def __getitem__(self, key):
        operator = AllOps.INDEX
        if isinstance(key, slice):
            key = slice_to_SLICE(key, self._dtype.bit_length())

        k = (operator, key)
        try:
            return self._usedOps[k]
        except KeyError:
            o = tv(self).__getitem__(self, key)
            self._usedOps[k] = o
            return o

    def _concat(self, *operands):
        return self.naryOp(AllOps.CONCAT, tv(self)._concat, *operands)

    def _ternary(self, ifTrue, ifFalse):
        return self.naryOp(AllOps.TERNARY, tv(self)._ternary, ifTrue, ifFalse)

    def _getIndexCascade(self):
        """
        Find out if this signal is something indexed
        """
        try:
            # now I am result of the index  xxx[xx] <= source
            # get index op
            d = self.singleDriver()
            try:
                op = d.operator
            except AttributeError:
                return

            if op == AllOps.INDEX:
                # get signal on which is index applied
                indexedOn = d.operands[0]
                if isinstance(indexedOn, RtlSignalBase):
                    # [TODO] multidimensional indexing
                    return indexedOn, [d.operands[1]]
                else:
                    raise Exception(
                        "can not drive static value %r" % indexedOn)

        except MultipleDriversExc:
            pass

    def __call__(self, source) -> List[Assignment]:
        """
        Create assignment to this signal

        :attention: it is not call of function it is operator of assignment
        :return: list of assignments
        """
        if isinstance(source, InterfaceBase):
            assert source._isAccessible
            source = source._sig

        if source is None:
            source = self._dtype.fromPy(None)
        else:
            source = toHVal(source, suggestedType=self._dtype)
            err = False
            try:
                source = source._auto_cast(self._dtype)
            except TypeConversionErr:
                err = True
            if err:
                raise TypeConversionErr(
                    ("Can not connect %r (of type %r) to %r "
                     "(of type %r) due type incompatibility")
                    % (source, source._dtype, self, self._dtype))

        tmp = self._getIndexCascade()
        if tmp:
            mainSig, indexCascade = tmp
            self = mainSig
        else:
            indexCascade = None

        # self = self._tryMyIndexToEndpoint()
        a = Assignment(source, self, indexCascade)

        self.drivers.append(a)
        if not isinstance(source, Value):
            source.endpoints.append(a)

        return [a]

    def __int__(self):
        if not self._const:
            raise TypeError("Int value of signal can be evaluated"
                            " because it is not constant expression:", self)
        else:
            return int(self._val)

    def __bool__(self):
        if not self._const:
            raise TypeError("Bool value of signal can be evaluated"
                            " because it is not constant expression:", self)
        else:
            return bool(self._val)

    def _isFullVld(self):
        return self._const and self._val._isFullVld()



from hwt.hdl.types.arrayVal import HArrayVal
from hwt.hdl.value import Value


class SliceVal(HArrayVal):
    """
    Value class for Slice type
    """

    def _isFullVld(self):
        return self.val[0]._isFullVld() and self.val[1]._isFullVld()

    def _size(self):
        """
        :return: how many bits is this slice selecting
        """
        assert isinstance(self, Value)
        return self.val[0].val - self.val[1].val



from copy import copy

from hwt.hdl.constants import INTF_DIRECTION, DIRECTION
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BIT
from hwt.hdl.types.struct import HStruct
from hwt.pyUtils.arrayQuery import single
from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import walkPhysInterfaces
from hwt.synthesizer.interfaceLevel.interfaceUtils.proxy import InterfaceProxy


def getClk(unit):
    try:
        return unit.clk
    except AttributeError:
        pass

    raise IntfLvlConfErr("Can not find clock on unit %r" % (unit,))


def getRst(unit):
    try:
        return unit.rst
    except AttributeError:
        pass

    try:
        return unit.rst_n
    except AttributeError:
        pass

    raise IntfLvlConfErr("Can not find clock on unit %r" % (unit,))


def getSignalName(sig):
    try:
        return sig._name
    except AttributeError:
        pass
    return sig.name


class UnitImplHelpers(object):
    def _reg(self, name, dtype=BIT, defVal=None):
        """
        Create register in this unit

        :param defVal: default value of this register, if this value is specified
            reset of this component is used
            (unit has to have single interface of class Rst or Rst_n)
        """
        clk = getClk(self)

        if defVal is None:
            # if no value is specified reset is not required
            rst = None
        else:
            rst = getRst(self)._sig

        if isinstance(dtype, HStruct):
            if defVal is not None:
                raise NotImplementedError()
            container = dtype.fromPy(None)
            for f in dtype.fields:
                if f.name is not None:
                    r = self._reg("%s_%s" % (name, f.name), f.dtype)
                    setattr(container, f.name, r)

            return container

        return self._cntx.sig(name,
                              dtype=dtype,
                              clk=clk._sig,
                              syncRst=rst,
                              defVal=defVal)

    def _sig(self, name, dtype=BIT, defVal=None):
        """
        Create signal in this unit
        """
        if isinstance(dtype, HStruct):
            if defVal is not None:
                raise NotImplementedError()
            container = dtype.fromPy(None)
            for f in dtype.fields:
                if f.name is not None:
                    r = self._sig("%s_%s" % (name, f.name), f.dtype)
                    setattr(container, f.name, r)

            return container

        return self._cntx.sig(name, dtype=dtype, defVal=defVal)

    def _cleanAsSubunit(self):
        """Disconnect internal signals so unit can be reused by parent unit"""
        for pi in self._entity.ports:
            pi.connectInternSig()
        for i in self._interfaces:
            i._clean()

    def _signalsForMyEntity(self, context, prefix):
        # generate for all ports of subunit signals in this context
        def lockTypeWidth(t):
            # [TODO] only read parameter instead of full evaluation
            # problem is that parametes should be theyr's values
            # (because this signals are for parent unit)
            if isinstance(t, Bits):
                t = copy(t)
                t.width = t.bit_length()
                return t
            else:
                return t

        for i in self._interfaces:
            if i._isExtern:
                i._signalsForInterface(context, prefix + i._NAME_SEPARATOR,
                                       typeTransform=lockTypeWidth)

    def _boundInterfacesToEntity(self, interfaces):
        externSignals = []
        inftToPortDict = {}

        for p in self._entity.ports:
            inftToPortDict[p._interface] = p

        for intf in self._interfaces:
            if intf._isExtern:
                for s in walkPhysInterfaces(intf):
                    externSignals.append(s)

        assert len(externSignals) == len(inftToPortDict.keys())

        for s in externSignals:
            self._boundIntfSignalToEntity(s, inftToPortDict)

    def _boundIntfSignalToEntity(self, interface, inftToPortDict):
        portItem = single(self._entity.ports, lambda x: x._interface == interface)
        interface._boundedEntityPort = portItem
        d = INTF_DIRECTION.asDirection(interface._direction)

        if d == DIRECTION.INOUT:
            portItem.direction = DIRECTION.INOUT

        if portItem.direction != d:
            # print(self._entity)
            # print(self._architecture)
            raise IntfLvlConfErr("Unit %s: Port %s does not have direction defined by interface %s, is %s should be %s" % 
                                 (self._name, portItem.name, repr(interface), portItem.direction, d))

    def _shareParamsWithPrefix(self, obj, prefix, paramNames):
        for name in paramNames:
            lp = getattr(obj, name)
            p = getattr(self, prefix + name)
            lp.set(p)

    def _updateParamsFrom(self, parent, exclude=None):
        """
        update all parameters which are defined on self from otherObj

        :param exclude: iterable of parameter on parent object which should be excluded
        """
        excluded = set()
        if exclude is not None:
            exclude = set(exclude)
        if isinstance(parent, InterfaceProxy):
            parent = parent._origIntf

        for parentP in parent._params:
            if exclude and parentP in exclude:
                excluded.add(parentP)
                continue

            name = parentP._scopes[parent][1]
            try:
                p = getattr(self, name)
            except AttributeError:
                continue
            p.set(parentP)

        if exclude is not None:
            assert excluded == exclude



from operator import eq, ne

from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.bitVal_bitOpsVldMask import vldMaskForOr, \
    vldMaskForAnd, vldMaskForXor
from hwt.hdl.types.bitVal_opReduce import tryReduceOr, tryReduceAnd, \
    tryReduceXor
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value, areValues
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


def boolLogOp__val(self, other, op, getVldFn):
    v = bool(op._evalFn(bool(self.val), (other.val)))
    return HBoolVal(v, BOOL,
                    getVldFn(self, other),
                    max(self.updateTime, other.updateTime))


def boolLogOp(self, other, op, getVldFn, reduceCheckFn):
    other = toHVal(other)

    iamVal = isinstance(self, Value)
    otherIsVal = isinstance(other, Value)

    if iamVal and otherIsVal:
        return boolLogOp__val(self, other, op, getVldFn)
    else:
        if otherIsVal:
            r = reduceCheckFn(self, other)
            if r is not None:
                return r

        elif iamVal:
            r = reduceCheckFn(other, self)
            if r is not None:
                return r

        return Operator.withRes(op, [self, other._auto_cast(BOOL)], BOOL)


def boolCmpOp__val(self, other, op, evalFn):
    v = evalFn(bool(self.val), bool(other.val)) and (
        self.vldMask == other.vldMask == 1)
    return HBoolVal(v, BOOL,
                    self.vldMask & other.vldMask,
                    max(self.updateTime, other.updateTime))


def boolCmpOp(self, other, op, evalFn=None):
    other = toHVal(other)
    if evalFn is None:
        evalFn = op._evalFn

    if areValues(self, other):
        return boolCmpOp__val(self, other, op, evalFn)
    else:
        return Operator.withRes(op, [self, other._auto_cast(BOOL)], BOOL)


class HBoolVal(Value):

    def _isFullVld(self):
        return bool(self.vldMask)

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        :param val: value of python type bool or None
        :param typeObj: instance of HdlType
        :param vldMask: None vldMask is resolved from val,
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        vld = int(val is not None)
        if not vld:
            assert vldMask is None or vldMask == 0
            val = False
        else:
            if vldMask == 0:
                val = False
                vld = 0
            else:
                val = bool(val)

        return cls(val, typeObj, vld)

    def _eq__val(self, other):
        return boolCmpOp__val(self, other, AllOps.EQ, eq)

    def _eq(self, other):
        return boolCmpOp(self, other, AllOps.EQ, evalFn=eq)

    def _ne__val(self, other):
        return boolCmpOp__val(self, other, AllOps.NEQ, ne)

    def __ne__(self, other):
        return boolCmpOp(self, other, AllOps.NEQ)

    def _invert__val(self):
        v = self.clone()
        v.val = not v.val
        return v

    def __invert__(self):
        if isinstance(self, Value):
            return self._invert__val()
        else:
            try:
                # double negation
                d = self.singleDriver()
                if isinstance(d, Operator) and d.operator == AllOps.NOT:
                    return d.operands[0]
            except MultipleDriversExc:
                pass
            return Operator.withRes(AllOps.NOT, [self], BOOL)

    def _ternary__val(self, ifTrue, ifFalse):
        if self.val:
            res = ifTrue.clone()
        else:
            res = ifFalse.clone()

        if not self.vldMask:
            res.vldMask = 0

        return res

    def _ternary(self, ifTrue, ifFalse):
        ifTrue = toHVal(ifTrue)
        ifFalse = toHVal(ifFalse)

        if isinstance(self, Value):
            return self._ternary__val(ifTrue, ifFalse)
        else:
            return Operator.withRes(AllOps.TERNARY, [self, ifTrue, ifFalse], ifTrue._dtype)

    # logic
    def _and__val(self, other):
        return boolLogOp__val(self, other, AllOps.AND, vldMaskForAnd)

    def __and__(self, other):
        return boolLogOp(self, other, AllOps.AND, vldMaskForAnd, tryReduceAnd)

    def _or__val(self, other):
        return boolLogOp__val(self, other, AllOps.OR, vldMaskForOr)

    def __or__(self, other):
        return boolLogOp(self, other, AllOps.OR, vldMaskForOr, tryReduceOr)

    def _xor__val(self, other):
        return boolLogOp__val(self, other, AllOps.OR, vldMaskForOr)

    def __xor__(self, other):
        return boolLogOp(self, other, AllOps.XOR, vldMaskForXor, tryReduceXor)



from hwt.hdl.types.bits import Bits
from hwt.serializer.verilog.utils import SIGNAL_TYPE


class VerilogSerializer_types():
    @classmethod
    def HdlType_bool(cls, typ, ctx, declaration=False):
        assert not declaration
        return ""

    @classmethod
    def HdlType_bits(cls, typ, ctx, declaration=False):
        isVector = typ.forceVector or typ.bit_length() > 1
        nameBuff = []
        sigType = ctx.signalType
        if sigType is SIGNAL_TYPE.PORT:
            pass
        elif sigType is SIGNAL_TYPE.REG:
            nameBuff.append("reg")
        elif sigType is SIGNAL_TYPE.WIRE:
            nameBuff.append("wire")
        else:
            raise NotImplementedError()

        if typ.signed:
            nameBuff.append("signed")

        w = typ.width
        if not isVector:
            pass
        elif isinstance(w, int):
            nameBuff.append("[%d:0]" % (w - 1))
        else:
            nameBuff.append("[%s- 1:0]" % cls.Value(w, ctx))

        return " ".join(nameBuff)

    @classmethod
    def HdlType_enum(cls, typ, ctx, declaration=False):
        if declaration:
            raise TypeError(
                "Verilog does not have enum types, hwt uses Bits instead")
        else:
            valueCnt = len(typ._allValues)
            return cls.HdlType_bits(Bits(valueCnt.bit_length()), ctx,
                                    declaration=declaration)

    @classmethod
    def HdlType_int(cls, typ, ctx, declaration=False):
        if ctx.signalType is SIGNAL_TYPE.PORT:
            return ""
        else:
            return "int"




class ConstCache(object):
    """
    Container of constants for serializer.
    Used to extract constants as constant variables.
    """

    def __init__(self, nameCheckFn):
        self.nameCheckFn = nameCheckFn

        # {value:usedName}
        self._cache = {}

    def getConstName(self, val):
        try:
            return self._cache[val]
        except KeyError:
            if isinstance(val.val, int):
                name = "const_%d_" % val.val
            else:
                name = "const_"

            c = self.nameCheckFn(name, val)
            self._cache[val] = c
            return c



from hwt.hdl.types.defs import INT, SLICE
from hwt.hdl.types.slice import Slice
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value


def slice_to_SLICE(sliceVals, width):
    """convert python slice to value of SLICE hdl type"""
    if sliceVals.step is not None:
        raise NotImplementedError()

    start = sliceVals.start
    stop = sliceVals.stop

    if sliceVals.start is None:
        start = INT.fromPy(width)
    else:
        start = toHVal(sliceVals.start)

    if sliceVals.stop is None:
        stop = INT.fromPy(0)
    else:
        stop = toHVal(sliceVals.stop)

    startIsVal = isinstance(start, Value)
    stopIsVal = isinstance(stop, Value)

    indexesAreValues = startIsVal and stopIsVal
    if indexesAreValues:
        updateTime = max(start.updateTime, stop.updateTime)
    else:
        updateTime = -1

    return Slice.getValueCls()((start, stop), SLICE, 1, updateTime)



from hwt.hdl.value import Value, areValues
from hwt.hdl.types.defs import BOOL, INT, SLICE
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.types.integer import Integer
from operator import pow, eq

BoolVal = BOOL.getValueCls()
SliceVal = SLICE.getValueCls()


def intOp__val(self, other, resT, evalFn):
    v = evalFn(self.val, other.val)
    vldMask = int(self.vldMask and other.vldMask)
    updateTime = max(self.updateTime, other.updateTime)
    return resT.getValueCls()(v, resT, vldMask, updateTime)


def intOp(self, other, op, resT, evalFn=None):
    if evalFn is None:
        evalFn = op._evalFn

    other = toHVal(other)._auto_cast(INT)
    if areValues(self, other):
        return intOp__val(self, other, resT, evalFn)
    else:
        return Operator.withRes(op, [self, other], resT)


def intAritmeticOp(self, other, op):
    return intOp(self, other, op, INT)


def intCmpOp(self, other, op, evalFn=None):
    return intOp(self, other, op, BOOL, evalFn=evalFn)


class IntegerVal(Value):

    def _isFullVld(self):
        return bool(self.vldMask)

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        :param val: value of python type int or None
        :param typeObj: instance of Integer
        :param vldMask: None vldMask is resolved from val,
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        assert isinstance(typeObj, Integer)
        vld = int(val is not None)
        if not vld:
            assert vldMask is None or vldMask == 0
            val = 0
        else:
            if vldMask == 0:
                val = False
                vld = 0
            else:
                val = int(val)

        return cls(val, typeObj, vld)

    # arithmetic
    def _neg__val(self):
        v = self.clone()
        v.val = -self.val
        return v

    def __neg__(self):
        if isinstance(self, Value):
            return self._neg__val()
        else:
            return Operator.withRes(AllOps.UN_MINUS, [self], INT)

    def __add__(self, other):
        return intAritmeticOp(self, other, AllOps.ADD)

    def __sub__(self, other):
        return intAritmeticOp(self, other, AllOps.SUB)

    def __mul__(self, other):
        return intAritmeticOp(self, other, AllOps.MUL)

    def _pow(self, other):
        return intOp(self, other, AllOps.POW, INT, pow)

    def __floordiv__(self, other):
        return intAritmeticOp(self, other, AllOps.DIV)

    def _downto__val(self, other):
        vldMask = int(self.vldMask and other.vldMask)
        updateTime = max(self.updateTime, other.updateTime)
        return SliceVal((self, other), SLICE, vldMask, updateTime)

    def _downto(self, other):
        other = toHVal(other)._auto_cast(INT)
        if areValues(self, other):
            return self._downto__val(other)
        else:
            return Operator.withRes(AllOps.DOWNTO, [self, other], SLICE)

    # comparisons
    def _eq(self, other):
        return intCmpOp(self, other, AllOps.EQ, eq)

    def __ne__(self, other):
        return intCmpOp(self, other, AllOps.NEQ)

    def __le__(self, other):
        return intCmpOp(self, other, AllOps.LE)

    def __lt__(self, other):
        return intCmpOp(self, other, AllOps.LT)

    def __ge__(self, other):
        return intCmpOp(self, other, AllOps.GE)

    def __gt__(self, other):
        return intCmpOp(self, other, AllOps.GT)




class LangueKeyword(object):
    pass


class NameOccupiedErr(Exception):
    def __init__(self, usedOn):
        self.usedOn = usedOn


class NameScopeItem(dict):
    """
    if name is discovered in scope it is converted to name_id
    where id is sequential number for prefix name\_
    """

    def __init__(self, myLvl):
        super().__init__()
        self.myLvl = myLvl

        # some names are specified just as prefix and serializer
        # should resolve correct name for object
        # this happens for most of generated objects
        self.cntrsForPrefixNames = {}

    def getChild(self, parent):
        try:
            return parent[self.myLvl + 1]
        except IndexError:
            return None

    def getParent(self, parent):

        i = self.myLvl - 1
        if i < 0:
            return None
        else:
            return parent[self.myLvl - 1]

    def __incrPrefixCntrsForChilds(self, prefix, currentVal, parent):
        # [TODO] check if new name is not defined in any direction
        currentVal += 1
        self.cntrsForPrefixNames[prefix] = currentVal
        ch = self.getChild(parent)
        while ch:
            if prefix in ch.cntrsForPrefixNames:
                ch.cntrsForPrefixNames[prefix] = currentVal
                ch = ch.getChild(parent)
            else:
                # prefix is not registered at any child
                break

        usableName = prefix + str(currentVal)
        return usableName

    def __registerName(self, name, obj, parent):
        # search if name is already defined on me and parents
        actual = self
        o = None

        if parent.ignorecase:
            _name = name.lower()
        else:
            _name = name

        while actual is not None:
            try:
                o = actual[_name]
            except KeyError:
                actual = actual.getParent(parent)
                continue
            break

        if o is None or o is obj:
            # we can use use the name, because it is not used
            self[_name] = obj
        else:
            raise NameOccupiedErr(o)

    def getUsableName(self, suggestedName, obj, parent):
        if not suggestedName.endswith("_"):
            try:
                self.__registerName(suggestedName, obj, parent)
                return suggestedName
            except NameOccupiedErr:
                suggestedName += "_"

        actual = self
        try:
            cntrVal = actual.cntrsForPrefixNames[suggestedName]
        except KeyError:
            cntrVal = -1

        # setup for me and propagate to children
        usableName = self.__incrPrefixCntrsForChilds(
            suggestedName, cntrVal, parent)
        self.__registerName(usableName, obj, parent)
        return usableName


class NameScope(list):
    """
    Scope of used names in hdl
    """

    def __init__(self, ignorecase):
        super().__init__()
        self.ignorecase = ignorecase

    def fork(self, lvl):
        f = self.__class__(self.ignorecase)
        for i in range(lvl):
            f.append(self[i])
        return f

    def setLevel(self, lvl):
        """
        Trim or extend scope
        lvl = 1 -> only one scope (global)
        """
        while len(self) != lvl:
            if len(self) > lvl:
                self.pop()
            else:
                self.append(NameScopeItem(len(self)))

    def checkedName(self, actualName, actualObj, isGlobal=False):
        if isGlobal:
            return self[0].getUsableName(actualName, actualObj, self)
        else:
            return self[-1].getUsableName(actualName, actualObj, self)



import sys

from hwt.code import connect
from hwt.hdl.architecture import Architecture
from hwt.hdl.constants import INTF_DIRECTION
from hwt.hdl.entity import Entity
from hwt.hdl.process import HWProcess
from hwt.hdl.statements import WaitStm
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bool import HBool
from hwt.hdl.types.defs import BIT
from hwt.hdl.types.enum import HEnum
from hwt.serializer.vhdl.serializer import VhdlSerializer
from hwt.simulator.hdlSimConfig import HdlSimConfig
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import \
    walkPhysInterfaces
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.rtlLevel.netlist import RtlNetlist
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


def makeTestbenchTemplate(unit: Unit, name: str=None):
    """
    :param unit: synthesized unit
    :return: (entity, arch, context) of testbench
    """
    if name is None:
        name = unit._name + "_tb"

    entity = Entity(name)
    arch = Architecture(entity)

    arch.components.append(unit._entity)
    arch.componentInstances.append(unit._entity)

    nl = RtlNetlist()
    ctx = {}
    for p in unit._entity.ports:
        t = p._dtype
        if isinstance(t, Bits) and not t == BIT:
            t = Bits(t.bit_length(), signed=t.signed,
                     forceVector=t.forceVector)
        s = RtlSignal(nl, p.name, t, t.fromPy(0))
        ctx[p._interface] = s
        p.connectSig(s)

    arch.variables.extend(ctx.values())

    return entity, arch, ctx


def mkDriverProc(intf, tbCtx):
    d = HWProcess(intf._sigInside.name + "_driver", [], set(), set(), set())
    d.actualTime = -1
    d.driverFor = tbCtx[intf]

    return d


class HdlSimConfigVhdlTestbench(HdlSimConfig):
    supported_type_classes = (HBool, Bits, HEnum)

    def __init__(self, top):
        self.top = top
        self.logPropagation = None
        self.logApplyingValues = None

        # unit :  signal | unit
        # signal : None
        self.registered = {}

    def beforeSim(self, simulator, synthesisedUnit):
        """
        This method is called before first step of simulation.
        """
        top = self.top
        self.tbEnt, self.tbArch, self.tbCtx = makeTestbenchTemplate(top)

        def reg(sigIntf):
            """register interface and create diver process for it"""
            proc = mkDriverProc(sigIntf, self.tbCtx)
            self.registered[sigIntf._sigInside] = proc
            self.tbArch.processes.append(proc)

        for s in walkPhysInterfaces(top):
            if s._direction is INTF_DIRECTION.SLAVE\
                    and isinstance(s._dtype, self.supported_type_classes):
                reg(s)

    def logChange(self, nowTime, sig, nextVal):
        """
        This method is called for every value change of any signal.
        """
        try:
            hwProc = self.registered[sig]
        except KeyError:
            # not every signal has to be registered
            return

        if hwProc.actualTime < nowTime:
            a = hwProc.actualTime
            if a < 0:
                a = 0
            delay = int(nowTime - a)
            if delay > 0:
                hwProc.statements.append(
                    WaitStm(int(delay) // 1000)
                )
                hwProc.actualTime = nowTime

        try:
            # SimBits type does not have forceVector flag,
            # but serializer requires it
            nextVal._dtype.forceVector = hwProc.driverFor._dtype.forceVector
        except AttributeError:
            pass

        hwProc.statements.extend(
            connect(nextVal, hwProc.driverFor)
        )

    def dump(self, dumpFile=sys.stdout):
        for proc in self.tbArch.processes:
            proc.statements.append(WaitStm(None))

        hasToBeOpened = isinstance(dumpFile, str)
        if hasToBeOpened:
            _dumpFile = open(dumpFile, 'w')
        else:
            _dumpFile = dumpFile

        ctx = VhdlSerializer.getBaseContext()
        ctx.scope.setLevel(2)
        VhdlSerializer.Entity_prepare(self.tbArch.components[0], ctx)
        ctx.scope.setLevel(1)

        _dumpFile.write(VhdlSerializer.formater(
            VhdlSerializer.Entity(self.tbEnt, ctx)))
        _dumpFile.write(VhdlSerializer.formater(
            VhdlSerializer.Architecture(self.tbArch, ctx)))



from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdl.hdlObject import HdlObject


class SignalItem(HdlObject):
    """
    basic hdl signal used to design circuits
    """

    def __init__(self, name, dtype, defaultVal=None, virtualOnly=False):
        """
        :param name: name for better orientation in netlists
            (used only in serialization)
        :param dtype: data type of this signal
        :param defaultVal: value for initialization
        :param virtualOnly: flag indicates that this assignments is only
            virtual and should not be added into
            netlist, because it is only for internal notation
        """
        self.name = name
        self._dtype = dtype
        self.virtualOnly = virtualOnly
        if defaultVal is None:
            defaultVal = dtype.fromPy(None)
        self.defaultVal = defaultVal
        self._setDefValue()

    def _setDefValue(self):
        v = self.defaultVal
        if isinstance(v, RtlSignalBase):
            v = v.staticEval()

        self._val = v.clone()
        self._oldVal = self._val.clone()
        self._oldVal.vldMask = 0



VERILOG_KEYWORDS = [
    "always"
    "and",
    "assign",
    "begin",
    "buf",
    "bufif0",
    "bufif1",
    "case",
    "casex",
    "casez",
    "cmos",
    "deassign",
    "default",
    "defparam",
    "disable",
    "edge",
    "else",
    "end",
    "endcase",
    "endfunction",
    "endmodule",
    "endprimitive",
    "endspecify",
    "endtable",
    "endtask",
    "event",
    "for",
    "force",
    "forever",
    "fork",
    "function",
    "highz0",
    "highz1",
    "if",
    "initial",
    "inout",
    "input",
    "integer",
    "join",
    "large",
    "macromodule",
    "medium",
    "module",
    "nand",
    "negedge",
    "nmos",
    "nor",
    "not",
    "notif0",
    "notif1",
    "or",
    "output",
    "pmos",
    "posedge",
    "primitive",
    "pull0",
    "pull1",
    "pulldown",
    "pullup",
    "rcmos",
    "reg",
    "release",
    "repeat",
    "rnmos",
    "rpmos",
    "rtran",
    "rtranif0",
    "rtranif1",
    "scalared",
    "small",
    "specify",
    "specparam",
    "strong0",
    "strong1",
    "supply0",
    "supply1",
    "table",
    "task",
    "time",
    "tran",
    "tranif0",
    "tranif1",
    "tri",
    "tri0",
    "tri1",
    "triand",
    "trior",
    "vectored",
    "wait",
    "wand",
    "weak0",
    "weak1",
    "while",
    "wire",
    "wor",
    "xnor",
    "xor",
]



from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.interfaceUtils.proxy import InterfaceProxy


class InterfaceArray():
    """
    Interface with multiplied widths of signals to create array of interfaces

    :ivar _arrayElemCache: list of elements in this interface array
    """
    def __init__(self):
        self._arrayElemCache = None

    def __len__(self):
        return int(self._asArraySize)

    def _initArrayItems(self):
        "instantiate my items into _arrayElemCache"
        self._arrayElemCache = []
        wm = int(self._widthMultiplier) // int(self._asArraySize)
        for index in range(len(self)):
            e = InterfaceProxy(self, index, index, None, wm)
            self._arrayElemCache.append(e)

    def _isInterfaceArray(self):
        return self._asArraySize is not None

    def __getitem__(self, key):
        try:
            return self._arrayElemCache[key]
        except IndexError as e:
            if not self._isInterfaceArray():
                raise IntfLvlConfErr("interface %s is not array and can not be indexed on" % self._name)
            else:
                raise e
        except TypeError:
            raise TypeError("Interface is not an array")



from hwt.bitmask import mask
from hwt.hdl.operator import Operator
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.integer import Integer
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value, areValues


BoolVal = BOOL.getValueCls()


def signFix(val, width):
    if val > 0:
        msb = 1 << (width - 1)
        if val & msb:
            val -= mask(width) + 1
    return val


def bitsCmp__val(self, other, op, evalFn):
    ot = other._dtype

    w = self._dtype._widthVal
    assert w == ot._widthVal, "%d, %d" % (w, ot._widthVal)

    vld = self.vldMask & other.vldMask
    _vld = vld == mask(w)
    res = evalFn(self.val, other.val) and _vld
    updateTime = max(self.updateTime, other.updateTime)

    return BoolVal(res, BOOL, int(_vld), updateTime)


def bitsCmp(self, other, op, evalFn=None):
    """
    :attention: If other is Bool signal convert this to bool (not ideal,
        due VHDL event operator)
    """
    other = toHVal(other)
    t = self._dtype
    ot = other._dtype

    iamVal = isinstance(self, Value)
    otherIsVal = isinstance(other, Value)

    if evalFn is None:
        evalFn = op._evalFn

    if iamVal and otherIsVal:
        if ot == BOOL:
            self = self._auto_cast(BOOL)
        elif t == ot:
            pass
        elif isinstance(ot, Integer):
            other = other._auto_cast(t)
        else:
            raise TypeError("Values of types (%r, %r) are not comparable" % (
                self._dtype, other._dtype))

        return bitsCmp__val(self, other, op, evalFn)
    else:
        if ot == BOOL:
            self = self._auto_cast(BOOL)
        elif t == ot:
            pass
        elif isinstance(ot, Integer):
            other = other._auto_cast(self._dtype)
        else:
            raise TypeError("Values of types (%r, %r) are not comparable" % (
                self._dtype, other._dtype))

        return Operator.withRes(op, [self, other], BOOL)


def bitsBitOp__val(self, other, op, getVldFn):
    w = self._dtype.bit_length()
    assert w == other._dtype.bit_length()

    vld = getVldFn(self, other)
    res = op._evalFn(self.val, other.val) & vld
    updateTime = max(self.updateTime, other.updateTime)
    if self._dtype.signed:
        res = signFix(res, w)

    return self.__class__(res, self._dtype, vld, updateTime)


def bitsBitOp(self, other, op, getVldFn, reduceCheckFn):
    """
    :attention: If other is Bool signal, convert this to bool
        (not ideal, due VHDL event operator)
    """
    other = toHVal(other)

    iamVal = isinstance(self, Value)
    otherIsVal = isinstance(other, Value)

    if iamVal and otherIsVal:
        other = other._auto_cast(self._dtype)
        return bitsBitOp__val(self, other, op, getVldFn)
    else:
        if other._dtype == BOOL:
            self = self._auto_cast(BOOL)
            return op._evalFn(self, other)
        elif self._dtype == other._dtype:
            pass
        else:
            raise TypeError("Can not apply operator %r (%r, %r)" %
                            (op, self._dtype, other._dtype))

        if otherIsVal:
            r = reduceCheckFn(self, other)
            if r is not None:
                return r

        elif iamVal:
            r = reduceCheckFn(other, self)
            if r is not None:
                return r

        return Operator.withRes(op, [self, other], self._dtype)


def bitsArithOp__val(self, other, op):
    v = self.clone()
    self_vld = self._isFullVld()
    other_vld = other._isFullVld()

    v.val = op._evalFn(self.val, other.val)

    w = v._dtype.bit_length()
    if self._dtype.signed:
        _v = v.val
        _max = mask(w - 1)
        _min = -_max - 1
        if _v > _max:
            _v = _min + (_v - _max - 1)
        elif _v < _min:
            _v = _max - (_v - _min + 1)

        v.val = _v
    else:
        v.val &= mask(w)

    if self_vld and other_vld:
        v.vldMask = mask(w)
    else:
        v.vldMask = 0

    v.updateTime = max(self.updateTime, other.updateTime)
    return v


def bitsArithOp(self, other, op):
    other = toHVal(other)
    assert isinstance(other._dtype, (Integer, Bits))
    if areValues(self, other):
        return bitsArithOp__val(self, other, op)
    else:
        resT = self._dtype
        if self._dtype.signed is None:
            self = self._unsigned()

        if isinstance(other._dtype, Bits):
            assert other._dtype.bit_length() == resT.bit_length(
            ), (op, other._dtype.bit_length(), resT.bit_length())
            other = other._convSign(self._dtype.signed)
        elif isinstance(other._dtype, Integer):
            pass
        else:
            raise TypeError("%r %r %r" % (self, op, other))

        o = Operator.withRes(op, [self, other], self._dtype)
        return o._auto_cast(resT)



from types import MethodType

from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase, InterfaceBase
from hwt.synthesizer.param import Param


def nameAvailabilityCheck(obj, propName, prop):
    """
    Check if not redefining property on obj
    """
    if getattr(obj, propName, None) is not None:
        raise IntfLvlConfErr("%r already has property %s old:%s new:%s" %
                             (obj, propName, repr(getattr(obj, propName)), prop))


class MakeParamsShared(object):
    """
    All newly added interfaces and units will share all parametes with unit
    specified in constructor of this object.
    """
    def __init__(self, unit, exclude=None):
        self.unit = unit
        self.exclude = exclude

    def __enter__(self):
        orig = self.unit._setAttrListener
        self.orig = orig
        exclude = self.exclude

        def MakeParamsSharedWrap(self, iName, i):
            if isinstance(i, (InterfaceBase, UnitBase)):
                i._updateParamsFrom(self, exclude=exclude)
            return orig(iName, i)

        self.unit._setAttrListener = MethodType(MakeParamsSharedWrap,
                                                self.unit)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.unit._setAttrListener = self.orig


class MakeClkRstAssociations(object):
    """
    All newly added interfaces will be associated with clk, rst
    specified in constructor of this object.
    """
    def __init__(self, unit, clk=None, rst=None):
        self.unit = unit
        self.clk = clk
        self.rst = rst

    def __enter__(self):
        orig = self.unit._setAttrListener
        self.orig = orig
        clk = self.clk
        rst = self.rst

        def MakeClkRstAssociationsWrap(self, iName, i):
            if isinstance(i, InterfaceBase):
                if clk is not None:
                    assert i._associatedClk is None
                    i._associatedClk = clk
                if rst is not None:
                    assert i._associatedRst is None
                    i._associatedRst = rst
            return orig(iName, i)

        self.unit._setAttrListener = MethodType(MakeClkRstAssociationsWrap,
                                                self.unit)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.unit._setAttrListener = self.orig


class PropDeclrCollector(object):
    """
    Collect properties of this object to containers by specified listeners
    """
    def _config(self):
        """
        Configure object parameters

        * setup all parameters on this object,
          use Param class instances to allow use of parametr inheritance
        * called in __init__ of class
        """
        pass

    def _declr(self):
        """
        declarations

        * do all declarations of externally accessible objects there (Interfaces)
        * is called after _config
        * _declr method is called after _config
        * if this object is Unit all interfaces are threaten as externally accessible interfaces
          if this object is Interface all subinterfaces are loaded
        """
        pass

    def _impl(self):
        """
        implementations

        * implement functionality of design there
        * called after _declr
        """
        pass

    def __setattr__(self, attr, value):
        """setattr with listener injector"""
        try:
            saListerner = self._setAttrListener
        except AttributeError:
            saListerner = None

        if saListerner:
            saListerner(attr, value)
        super().__setattr__(attr, value)

    # configuration phase
    def _loadConfig(self):
        if not hasattr(self, '_params'):
            self._params = []

        self._setAttrListener = self._paramCollector
        self._config()
        self._setAttrListener = None

    def _registerParameter(self, pName, parameter):
        """
        Register Param object on interface level object
        """
        nameAvailabilityCheck(self, pName, parameter)
        # resolve name in this scope
        try:
            hasName = parameter._name is not None
        except AttributeError:
            hasName = False
        if not hasName:
            parameter._name = pName
        # add name in this scope
        parameter._registerScope(pName, self)

        if parameter.hasGenericName:
            parameter.name = pName

        if parameter._parent is None:
            parameter._parent = self

        self._params.append(parameter)

    def _paramsShared(self, exclude=None):
        """
        Auto-propagate params by name to child components and interfaces
        Usage:

        .. code-block:: python

            with self._paramsShared():
                # your interfaces and unit which should share all params with "self" there

        :param exclude: params which should not be shared
        """
        return MakeParamsShared(self, exclude=exclude)

    def _associated(self, clk=None, rst=None):
        """
        associate newly added interfaces to "self" with selected clk, rst
        (if interface is not associated agents try to find clk/rst by _getAssociatedClk/_getAssociatedRst
        which will search for any clk/rst on parent recursively)
        Usage:

        .. code-block:: python

            with self._associated(clk=self.myClk, rst=self.myRst):
                self.myAxi = AxiStrem()
                # this interface is associated with myClk and myRst
                # simulation agents and component builders will use them


        :param exclude: params which should not be shared
        """
        return MakeClkRstAssociations(self, clk, rst)

    # declaration phase
    def _registerUnit(self, uName, unit):
        """
        Register unit object on interface level object
        """
        nameAvailabilityCheck(self, uName, unit)
        assert unit._parent is None
        unit._parent = self
        unit._name = uName
        self._units.append(unit)

    def _registerInterface(self, iName, intf):
        """
        Register interface object on interface level object
        """
        nameAvailabilityCheck(self, iName, intf)
        assert intf._parent is None
        intf._parent = self
        intf._name = iName
        intf._cntx = self._cntx
        self._interfaces.append(intf)

    def _declrCollector(self, name, prop):
        if name in ["_associatedClk", "_associatedRst"]:
            object.__setattr__(self, name, prop)
            return

        if isinstance(prop, InterfaceBase):
            self._registerInterface(name, prop)
        elif isinstance(prop, UnitBase):
            self._registerUnit(name, prop)

    def _registerArray(self, name, items):
        """
        Register array of items on interface level object
        """
        for i, itm in enumerate(items):
            setattr(self, name + str(i), itm)

    # implementation phase
    def _loadMyImplementations(self):
        self._setAttrListener = self._implCollector
        self._impl()
        self._setAttrListener = None

    def _registerUnitInImpl(self, uName, u):
        """
        :attention: unit has to be parametrized before it is registered
            (some components can change interface by parametrization)
        """
        self._registerUnit(uName, u)
        u._loadDeclarations()
        self._lazyLoaded.extend(u._toRtl(self._targetPlatform))
        u._signalsForMyEntity(self._cntx, "sig_" + uName)

    def _registerIntfInImpl(self, iName, i):
        """
        Register interface in implementation phase
        """
        raise NotImplementedError()

    def _paramCollector(self, pName, prop):
        if isinstance(prop, Param):
            self._registerParameter(pName, prop)

    def _implCollector(self, name, prop):
        if isinstance(prop, InterfaceBase):
            self._registerIntfInImpl(name, prop)
        elif isinstance(prop, UnitBase):
            self._registerUnitInImpl(name, prop)



from hwt.hdl.constants import INTF_DIRECTION
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.param import evalParam, Param


class IpCoreWrapper(Unit):
    """
    Class which creates wrapper which converts all incompatible parts of unit
    to ipcore compatible this means:

    * convert array interface to multiple interfaces

    original unit will be placed inside as subunit named baseUnit
    """

    def __init__(self, baseUnit):
        super(IpCoreWrapper, self).__init__()
        self._baseUnit = baseUnit

    def _copyParamsAndInterfaces(self):
        for p in self._baseUnit._params:
            myP = Param(evalParam(p))
            self._registerParameter(p.name, myP)
            object.__setattr__(self, myP.name, myP)

        origToWrapInfMap = {}

        for intf in self.baseUnit._interfaces:
            if intf._asArraySize is not None:
                # create array of interfaces instead of array interface
                for i in range(int(intf._asArraySize)):
                    myIntf = intf._clone()
                    name = intf._name + "_%d" % i

                    self._registerInterface(name, myIntf)
                    object.__setattr__(self, name, myIntf)

                    try:
                        ia = origToWrapInfMap[intf]
                    except KeyError:
                        ia = []
                        origToWrapInfMap[intf] = ia

                    ia.append(myIntf)

            else:
                # clone interface
                myIntf = intf._clone()

                self._registerInterface(intf._name, myIntf)
                object.__setattr__(self, intf._name, myIntf)

                origToWrapInfMap[intf] = myIntf

        for i in self._interfaces:
            self._loadInterface(i, True)

        return origToWrapInfMap

    def _getDefaultName(self):
        return self._baseUnit.__class__.__name__

    def _lazyLoadParamsAndInterfaces(self):
        self._cntx.params = self._buildParams()

        # prepare signals for interfaces
        for i in self._interfaces:
            assert i._isExtern
            signals = i._signalsForInterface(self._cntx)
            self._externInterf.extend(signals)

    def _connectBaseUnitToThisWrap(self, origToWrapInfMap):
        for baseIntf, wrapIntf in origToWrapInfMap.items():
            if baseIntf._direction is INTF_DIRECTION.MASTER:
                if isinstance(wrapIntf, list):
                    for i, _wrapIntf in enumerate(wrapIntf):
                        _wrapIntf(baseIntf[i])
                else:
                    wrapIntf(baseIntf)
            else:
                if isinstance(wrapIntf, list):
                    for i, _wrapIntf in enumerate(wrapIntf):
                        baseIntf[i](_wrapIntf)
                else:
                    baseIntf(wrapIntf)

    def _impl(self):
        self.baseUnit = self._baseUnit
        origToWrapInfMap = self._copyParamsAndInterfaces()
        self._lazyLoadParamsAndInterfaces()
        self._connectBaseUnitToThisWrap(origToWrapInfMap)




class SerializerException(Exception):
    pass


class UnsupportedEventOpErr(SerializerException):
    """
    Target HDL can not use event operator in this context, it usually
    has to be replaced by correct expression of sensitivity list
    """



from hwt.bitmask import mask
from hwt.hdl.operator import Operator
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, BIT
from hwt.hdl.value import Value
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.value import GenericSerializer_Value
from hwt.serializer.generic.indent import getIndent
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class VhdlSerializer_Value(GenericSerializer_Value):
    @classmethod
    def condAsHdl(cls, cond, forceBool, ctx):
        if isinstance(cond, RtlSignalBase):
            cond = [cond]
        else:
            cond = list(cond)
        if len(cond) == 1:
            c = cond[0]
            if not forceBool or c._dtype == BOOL:
                return cls.asHdl(c, ctx)
            elif c._dtype == BIT:
                return cls.asHdl(c._eq(1), ctx)
            elif isinstance(c._dtype, Bits):
                return cls.asHdl(c != 0, ctx)
            else:
                raise NotImplementedError()
        else:
            return " AND ".join(map(lambda x: cls.condAsHdl(x, forceBool, ctx),
                                    cond))

    @classmethod
    def SignalItem(cls, si, ctx, declaration=False):
        if declaration:
            v = si.defaultVal
            if si.virtualOnly:
                prefix = "VARIABLE"
            elif si.drivers:
                prefix = "SIGNAL"
            elif si.endpoints or si.simSensProcs:
                prefix = "CONSTANT"
                if not v.vldMask:
                    raise SerializerException(
                        "Signal %s is constant and has undefined value"
                        % si.name)
            else:
                raise SerializerException(
                    "Signal %s should be declared but it is not used"
                    % si.name)

            s = "%s%s %s: %s" % (getIndent(ctx.indent),
                                 prefix, si.name, cls.HdlType(si._dtype, ctx))
            if isinstance(v, RtlSignalBase):
                return s + " := %s" % cls.asHdl(v, ctx)
            elif isinstance(v, Value):
                if si.defaultVal.vldMask:
                    return s + " := %s" % cls.Value(si.defaultVal, ctx)
                else:
                    return s
            else:
                raise NotImplementedError(v)

        else:
            if si.hidden and hasattr(si, "origin"):
                return cls.asHdl(si.origin, ctx)
            else:
                return si.name

    @classmethod
    def HEnumValAsHdl(cls, dtype, val, ctx):
        return '%s' % str(val.val)

    @classmethod
    def HArrayValAsHdl(cls, dtype, val, ctx):
        separator = ",\n" + getIndent(ctx.indent + 1)
        return "".join(["(",
                        separator.join([cls.Value(v, ctx) for v in val]),
                        ")"])

    @staticmethod
    def BitString_binary(v, width, vldMask=None):
        buff = ['"']
        for i in range(width - 1, -1, -1):
            mask = (1 << i)
            b = v & mask

            if vldMask & mask:
                s = "1" if b else "0"
            else:
                s = "X"
            buff.append(s)
        buff.append('"')
        return ''.join(buff)

    @classmethod
    def BitString(cls, v, width, vldMask=None):
        if vldMask is None:
            vldMask = mask(width)
        # if can be in hex
        if width % 4 == 0 and vldMask == (1 << width) - 1:
            return ('X"%0' + str(width // 4) + 'x"') % (v)
        else:  # else in binary
            return cls.BitString_binary(v, width, vldMask)

    @classmethod
    def BitLiteral(cls, v, vldMask):
        if vldMask:
            return "'%d'" % int(bool(v))
        else:
            return "'X'"

    @classmethod
    def sensitivityListItem(cls, item, ctx):
        if isinstance(item, Operator):
            item = item.operands[0]
        return cls.asHdl(item, ctx)

    @classmethod
    def SignedBitString(cls, v, width, forceVector, vldMask):
        if vldMask != mask(width):
            if forceVector or width > 1:
                v = cls.BitString(v, width, vldMask)
            else:
                v = cls.BitLiteral(v, width, vldMask)
        else:
            v = str(v)
        # [TODO] parametrized width
        return "TO_SIGNED(%s, %d)" % (v, width)

    @classmethod
    def UnsignedBitString(cls, v, width, forceVector, vldMask):
        if vldMask != mask(width):
            if forceVector or width > 1:
                v = cls.BitString(v, width, vldMask)
            else:
                v = cls.BitLiteral(v, width, vldMask)
        else:
            v = str(v)
        # [TODO] parametrized width
        return "TO_UNSIGNED(%s, %d)" % (v, width)

    @classmethod
    def Bool_valAsHdl(cls, dtype, val, ctx):
        return str(bool(val.val))

    @classmethod
    def Slice_valAsHdl(cls, dtype, val, ctx):
        upper = val.val[0]
        if isinstance(upper, Value):
            upper = upper - 1
            _format = "%s DOWNTO %s"
        else:
            _format = "%s-1 DOWNTO %s"

        return _format % (cls.Value(upper, ctx), cls.Value(val.val[1], ctx))

    @classmethod
    def String_valAsHdl(cls, dtype, val, ctx):
        return '"%s"' % str(val.val)



from hwt.hdl.value import Value
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.typeCast import toHVal


class StringVal(Value):
    """
    Value class for hdl String type
    """

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        :param val: python string or None
        :param typeObj: instance of String HdlType
        :param vldMask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        assert isinstance(val, str) or val is None
        vld = 0 if val is None else 1
        if not vld:
            assert vldMask is None or vldMask == 0
            val = ""
        else:
            if vldMask == 0:
                val = ""
                vld = 0

        return cls(val, typeObj, vld)

    def _eq__val(self, other):
        eq = self.val == other.val
        vld = int(self.vldMask and other.vldMask)
        updateTime = max(self.updateTime, other.updateTime)

        return BOOL.getValueCls()(eq, BOOL, vld, updateTime)

    def _eq(self, other):
        other = toHVal(other)
        if isinstance(other, Value):
            return self._eq__val(other)
        else:
            raise NotImplementedError()



from collections import deque

from hwt.hdl.constants import NOP
from hwt.simulator.agentBase import SyncAgentBase


class RdSyncedAgent(SyncAgentBase):
    """
    Simulation/verification agent for RdSynced interface
    """
    def __init__(self, intf, allowNoReset=True):
        super().__init__(intf, allowNoReset=allowNoReset)
        self.actualData = NOP
        self.data = deque()
        self._rd = self.getRd(intf)

    def getRd(self, intf):
        return intf.rd

    def isRd(self, readFn):
        return readFn(self._rd)

    def wrRd(self, writeFn, val):
        writeFn(val, self._rd)

    def setEnable_asMonitor(self, en, sim):
        super(RdSyncedAgent, self).setEnable_asMonitor(en, sim)
        if not en:
            self.wrRd(sim.write, 0)

    def monitor(self, sim):
        """Collect data from interface"""
        if self.notReset(sim) and self._enabled:
            self.wrRd(sim.write, 1)

            yield sim.waitOnCombUpdate()

            d = self.doRead(sim)
            self.data.append(d)
        else:
            self.wrRd(sim.write, 0)

    def doRead(self, sim):
        """extract data from interface"""
        return sim.read(self.intf.data)

    def doWrite(self, sim, data):
        """write data to interface"""
        sim.write(data, self.intf.data)

    def driver(self, sim):
        """Push data to interface"""
        r = sim.read
        if self.actualData is NOP and self.data:
            self.actualData = self.data.popleft()

        do = self.actualData is not NOP

        if do:
            self.doWrite(sim, self.actualData)
        else:
            self.doWrite(sim, None)

        en = self.notReset(sim) and self._enabled
        if not (en and do):
            return

        yield sim.waitOnCombUpdate()

        rd = self.isRd(r)
        if en:
            assert rd.vldMask, (
                ("%r: ready signal for interface %r is in invalid state,"
                 " this would cause desynchronization") %
                (sim.now, self.intf))
        if rd.val:
            if self._debugOutput is not None:
                self._debugOutput.write("%s, wrote, %d: %r\n" % (
                                           self.intf._getFullName(),
                                           sim.now, self.actualData))
            if self.data:
                self.actualData = self.data.popleft()
            else:
                self.actualData = NOP



from hwt.serializer.vhdl.serializer import VhdlSerializer


class VivadoTclExpressionSerializer(VhdlSerializer):
    # disabled because this code is not reachable in current implemetation
    @staticmethod
    def SignalItem(si, declaration=False):
        raise NotImplementedError()



from hwt.code import connect
from hwt.interfaces.std import Clk, Rst_n, Rst


def addClkRstn(obj):
    """
    Construct clk, rst_n signal on object (usually Unit/Interface instance)
    """
    obj.clk = Clk()
    obj.rst_n = Rst_n()


def addClkRst(obj):
    """
    Construct clk, rst signal on object (usually Unit/Interface instance)
    """
    obj.clk = Clk()
    obj.rst = Rst()


def _tryConnect(src, unit, intfName):
    """
    Try connect src to interface of specified name on unit.
    Ignore if interface is not present or if it already has driver.
    """
    try:
        dst = getattr(unit, intfName)
    except AttributeError:
        return
    if not dst._sig.drivers:
        connect(src, dst)


def propagateClk(obj):
    """
    Propagate "clk" clock signal to all subcomponents
    """
    clk = obj.clk
    for u in obj._units:
        _tryConnect(clk, u, 'clk')


def propagateClkRstn(obj):
    """
    Propagate "clk" clock and negative reset "rst_n" signal
    to all subcomponents
    """
    clk = obj.clk
    rst_n = obj.rst_n

    for u in obj._units:
        _tryConnect(clk, u, 'clk')
        _tryConnect(rst_n, u, 'rst_n')
        _tryConnect(~rst_n, u, 'rst')


def propagateClkRst(obj):
    """
    Propagate "clk" clock and reset "rst" signal to all subcomponents
    """
    clk = obj.clk
    rst = obj.rst

    for u in obj._units:
        _tryConnect(clk, u, 'clk')
        _tryConnect(~rst, u, 'rst_n')
        _tryConnect(rst, u, 'rst')



from functools import reduce
from math import ceil
from typing import Union

from hwt.hdl.typeShortcuts import vec
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.structUtils import walkFlattenFields
from hwt.hdl.value import Value
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.hdl.types.hdlType import HdlType


class BitWidthErr(Exception):
    """
    Wrong bit width of signal/value
    """


def fitTo_t(what: Union[RtlSignal, Value], where_t: HdlType,
            extend: bool=True, shrink: bool=True):
    """
    Slice signal "what" to fit in "where"
    or
    arithmetically (for signed by MSB / unsigned, vector with 0) extend
    "what" to same width as "where"

    little-endian impl.
    """

    whatWidth = what._dtype.bit_length()
    toWidth = where_t.bit_length()
    if toWidth == whatWidth:
        return what
    elif toWidth < whatWidth:
        # slice
        if not shrink:
            raise BitWidthErr()

        return what[toWidth:]
    else:
        if not extend:
            raise BitWidthErr()

        w = toWidth - whatWidth

        if what._dtype.signed:
            # signed extension
            msb = what[whatWidth - 1]
            ext = reduce(lambda a, b: a._concat(b), [msb for _ in range(w)])
        else:
            # 0 extend
            ext = vec(0, w)

        return ext._concat(what)


def fitTo(what: Union[RtlSignal, Value], where: Union[RtlSignal, Value],
          extend: bool=True, shrink: bool=True):
    return fitTo_t(what, where._dtype, extend, shrink)


class NotEnoughtBitsErr(Exception):
    """
    More bits is required for such an operation
    """


class BitWalker():
    """
    Walker which can walk chunks of bits on signals/values of all types

    :ivar sigOrVal: signal or value to iterate over
    :ivar fillup: flag that means that if there is not enought bits
        for last iterm fill it up with invalid bits (otherwise raise)
    """

    def __init__(self, sigOrVal: Union[RtlSignal, Value],
                 skipPadding: bool=True,
                 fillup: bool=False):
        """
        :param skipPadding: if true padding is skipped in dense types
        """
        self.it = walkFlattenFields(sigOrVal, skipPadding=skipPadding)
        self.fillup = fillup
        self.actuallyHave = 0
        self.actual = None
        self.actualOffset = 0

    def _get(self, numberOfBits: int, doCollect: bool):
        """
        :param numberOfBits: number of bits to get from actual possition
        :param doCollect: if False output is not collected just iterator moves
            in structure
        """
        if not isinstance(numberOfBits, int):
            numberOfBits = int(numberOfBits)

        while self.actuallyHave < numberOfBits:
            # accumulate while not has enought
            try:
                f = next(self.it)
            except StopIteration:
                if self.fillup and self.actual is not None:
                    break
                else:
                    raise NotEnoughtBitsErr()

            thisFieldLen = f._dtype.bit_length()
            if self.actual is None:
                if not doCollect and thisFieldLen <= numberOfBits:
                    numberOfBits -= thisFieldLen
                else:
                    self.actual = f
                    self.actuallyHave = thisFieldLen
            else:
                if not doCollect and self.actuallyHave < numberOfBits:
                    self.actuallyHave = thisFieldLen
                    self.actual = f
                else:
                    self.actuallyHave += thisFieldLen
                    self.actual = f._concat(self.actual)

        # slice out from actual
        actual = self.actual
        actualOffset = self.actualOffset

        if self.actuallyHave < numberOfBits:
            assert self.fillup
            if doCollect:
                t = self.actual._dtype
                fillupW = numberOfBits - self.actuallyHave
                padding_t = Bits(fillupW, signed=t.signed, negated=t.negated)
                padding = padding_t.fromPy(None)
                actual = padding._concat(actual)
            self.actuallyHave = 0

        # update about what was taken
        self.actuallyHave -= numberOfBits
        self.actualOffset += numberOfBits
        if self.actuallyHave == 0:
            self.actual = None
            self.actualOffset = 0

        if doCollect:
            if numberOfBits == 1:
                return actual[actualOffset]
            else:
                return actual[(actualOffset + numberOfBits):actualOffset]

    def get(self, numberOfBits: int) -> Union[RtlSignal, Value]:
        """
        :param numberOfBits: number of bits to get from actual possition
        :return: chunk of bits of specified size (instance of Value or RtlSignal)
        """
        return self._get(numberOfBits, True)

    def skip(self, numberOfBits: int) -> None:
        """
        Move this iterator without care about item

        :param numberOfBits: number of bits to get from actual possition
        """
        self._get(numberOfBits, False)

    def assertIsOnEnd(self):
        """
        Assert there is nothing left in this iterator
        """
        try:
            next(self.it)
        except StopIteration:
            return

        raise AssertionError("BitWalker there stil were some items")


def iterBits(sigOrVal: Union[RtlSignal, Value], bitsInOne: int=1,
             skipPadding: bool=True, fillup: bool=False):
    """
    Iterate over bits in vector

    :param sigOrVal: signal or value to iterate over
    :param bitsInOne: number of bits in one part
    :param skipPadding: if true padding is skipped in dense types
    """
    bw = BitWalker(sigOrVal, skipPadding, fillup)
    for _ in range(ceil(sigOrVal._dtype.bit_length() / bitsInOne)):
        yield bw.get(bitsInOne)

    bw.assertIsOnEnd()



from hwt.hdl.variables import SignalItem
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.constants import SIGNAL_TYPE
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.systemC.utils import systemCTypeOfSig


class SystemCSerializer_statements():

    @classmethod
    def _Assignment(cls, dst, typeOfDst, src, ctx):
        indent_str = getIndent(ctx.indent)

        dstStr = cls.asHdl(dst, ctx.forTarget())
        if typeOfDst == SIGNAL_TYPE.REG:
            fmt = "%s%s = %s;"
        else:
            fmt = "%s%s.write(%s);"

        return fmt % (indent_str, dstStr, cls.Value(src, ctx))

    @classmethod
    def Assignment(cls, a, ctx):
        dst = a.dst
        assert isinstance(dst, SignalItem)
        assert not dst.virtualOnly, "should not be required"

        typeOfDst = systemCTypeOfSig(dst)
        if a.indexes is not None:
            for i in a.indexes:
                dst = dst[i]

        if dst._dtype == a.src._dtype:
            return cls._Assignment(dst, typeOfDst, a.src, ctx)
        else:
            raise SerializerException("%r <= %r  is not valid assignment\n"
                                      " because types are different (%r; %r) "
                                      % (dst, a.src, dst._dtype, a.src._dtype))

    @classmethod
    def HWProcess(cls, proc, ctx):
        """
        Serialize HWProcess instance

        :param scope: name scope to prevent name collisions
        """
        body = proc.statements
        childCtx = ctx.withIndent()
        statemets = [cls.asHdl(s, childCtx) for s in body]
        proc.name = ctx.scope.checkedName(proc.name, proc)

        return cls.methodTmpl.render(
            indent=getIndent(ctx.indent),
            name=proc.name,
            statements=statemets
        )



# select = map, groupBy = itertools.groupby
from types import GeneratorType
from itertools import zip_longest
from math import inf


class DuplicitValueExc(Exception):
    """
    Exception which means that there are multiple items which this query
    selected but it should return only one
    """


class NoValueExc(Exception):
    """
    Exception which means that query did not selected any item
    """


def distinctBy(iterable, fn):
    """
    uniq operation with key selector
    """
    s = set()
    for i in iterable:
        r = fn(i)
        if r not in s:
            s.add(r)
            yield i


def single(iterable, fn):
    """
    Get value from iterable where fn(item) and check
    if there is not fn(other item)

    :raise DuplicitValueExc: when there are multiple items satisfying fn()
    :raise NoValueExc: when no value satisfying fn(item) found
    """
    found = False
    ret = None

    for i in iterable:
        if fn(i):
            if found:
                raise DuplicitValueExc(i)
            found = True
            ret = i

    if not found:
        raise NoValueExc()

    return ret


def arr_any(iterable, fn):
    """
    :return: True if fn(item) for any item else False
    """
    for item in iterable:
        if fn(item):
            return True
    return False


def arr_all(iterable, fn):
    """
    :return: True if fn(item) for all items in interable or iterable
        is empty else False
    """
    for item in iterable:
        if not fn(item):
            return False
    return True


def take(iterrable, howMay):
    """
    :return: generator of first n items from iterrable
    """
    assert howMay >= 0

    if not howMay:
        return

    last = howMay - 1
    for i, item in enumerate(iterrable):
        yield item
        if i == last:
            return


def where(iterable, fn):
    """
    :return: generator of items from iterable where fn(item)
    """
    for i in iterable:
        if fn(i):
            yield i


def iter_with_last(iterable):
    """
    :return: generator of tuples (isLastFlag, item)
    """
    # Ensure it's an iterator and get the first field
    iterable = iter(iterable)
    prev = next(iterable)
    for item in iterable:
        # Lag by one item so I know I'm not at the end
        yield False, prev
        prev = item
    # Last item
    yield True, prev


def groupedby(collection, fn):
    """
    same like itertools.groupby

    :note: This function does not needs initial sorting like itertools.groupby

    :attention: Order of pairs is not deterministic.
    """
    d = {}
    for item in collection:
        k = fn(item)
        try:
            arr = d[k]
        except KeyError:
            arr = []
            d[k] = arr
        arr.append(item)

    yield from d.items()


def flatten(iterables, level=inf):
    """
    Flatten nested lists, tuples, generators and maps

    :param level: maximum depth of flattening
    """
    if level >= 0 and isinstance(iterables, (list, tuple, GeneratorType,
                                             map, zip)):
        level -= 1
        for i in iterables:
            yield from flatten(i, level=level)
    else:
        yield iterables


def grouper(n, iterable, padvalue=None):
    """grouper(3, 'abcdefg', 'x') -->
       ('a','b','c'), ('d','e','f'), ('g','x','x')
    """
    return zip_longest(*[iter(iterable)] * n, fillvalue=padvalue)


def areSetsIntersets(setA, setB):
    """
    Check if intersection of sets is not empty
    """
    return any(x in setA for x in setB)


def balanced_reduce(arr, opFn):
    while len(arr) > 1:
        nextArr = []
        for a, b in grouper(2, arr):
            nextArr.append(opFn(a, b))
        arr = nextArr

    return arr[0]



from collections import deque

from hwt.hdl.constants import NOP
from hwt.simulator.agentBase import SyncAgentBase


class VldSyncedAgent(SyncAgentBase):
    def __init__(self, intf, allowNoReset=False):
        super(VldSyncedAgent, self).__init__(intf,
                                             allowNoReset=allowNoReset)
        self.data = deque()

    def doRead(self, s):
        return s.read(self.intf.data)

    def doWrite(self, s, data):
        s.write(data, self.intf.data)

    def setEnable_asDriver(self, en, sim):
        super(VldSyncedAgent, self).setEnable_asDriver(en, sim)
        if not en:
            self.wrVld(sim.write, 0)
            self._lastVld = 0

    def monitor(self, sim):
        intf = self.intf
        yield sim.waitOnCombUpdate()
        if self.notReset(sim):
            vld = sim.read(intf.vld)
            assert vld.vldMask, (
                ("valid signal for interface %r is in invalid state,"
                 " this would cause desynchronization in %d") %
                (self.intf, sim.now))
            if vld.val:
                d = self.doRead(sim)

                if self._debugOutput is not None:
                    self._debugOutput.write("%s, read, %d: %r\n" % (
                        self.intf._getFullName(),
                        sim.now, d))
                self.data.append(d)

    def driver(self, sim):
        intf = self.intf

        if self.data and self.notReset(sim):
            d = self.data.popleft()
            if d is NOP:
                self.doWrite(sim, None)
                sim.write(0, intf.vld)
            else:
                self.doWrite(sim, d)
                sim.write(1, intf.vld)
                if self._debugOutput is not None:
                    self._debugOutput.write("%s, wrote, %d: %r\n" % (
                        self.intf._getFullName(),
                        sim.now, self.actualData))

        else:
            self.doWrite(sim, None)
            sim.write(0, intf.vld)



from hwt.bitmask import mask
from hwt.hdl.types.bits import Bits
from hwt.simulator.types.simBitsConversions import convertSimBits__val
from typing import Union


__simBitsTCache = {}


def simBitsT(width: int, signed: Union[bool, None]):
    """
    Construct SimBitsT with cache
    """
    k = (width, signed)
    try:
        return __simBitsTCache[k]
    except KeyError:
        t = SimBitsT(width, signed)
        __simBitsTCache[k] = t
        return t


class SimBitsT(Bits):
    """
    Simplified Bits type for simulation purposes
    """

    def __init__(self, width, signed):
        self._widthVal = width
        self.signed = signed
        self._allMask = mask(self._widthVal)
        self.forceVector = False

    def __eq__(self, other):
        return isinstance(other, Bits) and other._widthVal == self._widthVal\
            and self.signed == other.signed

    def __hash__(self):
        return hash((self._widthVal, self.signed))

    def convert(self, sigOrVal, toType):
        return convertSimBits__val(self, sigOrVal, toType)

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        return "<SimBitsT, signed:%r, %dbits>" % (self.signed, self._widthVal)


SIM_BIT = SimBitsT(1, None)
